;--------------------------------
;TP-RTL2.INC	ho	01-MAR-94
;--------------------------------
;V1.00: 04BFh-07C9h
;V2.00: 04FCh-0806h	V2.01: 04F6h-0800h
;V3.00: 052Ch-083Ch	V3.01: 052Dh-083Dh

		IDENT	TP-RTL2.INC

;get data (from memory)
;----------------------
;(subroutines are provided only for non-trivial data types)

;get REAL number (to registers)
;E: pointer to REAL number in HL
;V: BC, DE, HL
;R: REAL number in registers BCDEHL (most significant -> least significant)
LoadReal:	LD	E,(HL)
		INC	HL
		LD	D,(HL)
		INC	HL
		PUSH	DE
		LD	E,(HL)
		INC	HL
		LD	D,(HL)
		INC	HL
		LD	C,(HL)
		INC	HL
		LD	B,(HL)
		POP	HL
		RET

;get string (copy to stack)
;E: HL = pointer to string
;V: AF, BC, DE, HL, IX, stack / R: BC=0
LoadString:	POP	IX		;take return address from stack
		EX	DE,HL
		LD	A,(DE)		;get string length
		LD	C,A
		LD	B,0
		CPL			;reserve space on stack
		LD	L,A		;by decreasing
		LD	H,0FFh		;the stackpointer
		ADD	HL,SP
		LD	SP,HL
		EX	DE,HL		;HL points to string, DE to stack
		INC	BC		;one more because of length byte
		LDIR			;copy string
		JP	(IX)		;return

;get constant string (copy to stack)
;E: (SP) = pointer to string
;V: AF, BC, DE, HL, stack / R: BC=0
;Compare with LoadString. A common subroutine could save a few bytes here.
LoadCString:	POP	DE		;return address is pointer to string
		LD	A,(DE)		;get length
		LD	C,A
		LD	B,0
		CPL			;reserve space on stack
		LD	L,A		;by decreasing
		LD	H,0FFh		;the stackpointer
		ADD	HL,SP
		LD	SP,HL
		EX	DE,HL
		INC	BC		;copy string. HL points after
		LDIR			;end of string, thats where
		JP	(HL)		;we continue execution

;get SET (copy to stack)
;E: pointer to partial SET in HL, length of partial SET in C,
;   offset of partial SET in SET in B
;V: AF, B, DE, HL, IX, stack
LoadSet:	POP	IX		;take return address from stack
		EX	DE,HL
		LD	HL,-32
		ADD	HL,SP		;reserve 32 byte
		LD	SP,HL		;on stack
		EX	DE,HL
		PUSH	BC		;write B zeros
		INC	B		;and copy C bytes from HL
		DEC	B
		JR	Z,LoadSet2
		XOR	A
LoadSet1:	LD	(DE),A
		INC	DE
		DJNZ	LoadSet1
LoadSet2:	LDIR
		POP	BC
		LD	A,32		;if there is some space left
		SUB	B		;fill the remaining bytes
		SUB	C		;with zeros
		JR	Z,LoadSet4
		LD	B,A
		XOR	A
LoadSet3:	LD	(DE),A
		INC	DE
		DJNZ	LoadSet3
LoadSet4:	JP	(IX)

;create a constant SET >>
;first create an empty SET, then add the single elements or subranges

;create an empty SET (on stack)
;E:- / V: AF, B, HL, IX, stack / R:-
EmptySet:	POP	IX
		LD	HL,-32		;reserve space for SET
		ADD	HL,SP
		LD	SP,HL
		LD	B,32		;clear SET
		XOR	A
EmptySet1:	LD	(HL),A
		INC	HL
		DJNZ	EmptySet1
		JP	(IX)

;add a single element to SET (which is on stack)
;E: element in HL (max. 255)
;V: AF, B, HL, IX
;R: modified SET on stack
ElemToSet:	POP	IX		;save return address
		LD	B,L
		CALL	PtrMskSet	;calculate pointer and mask
ToSet1:		OR	(HL)		;set bit
		LD	(HL),A
ToSet2:		JP	(IX)		;return

;add a subrange (FROM..TO) to SET (which is on stack)
;E: TO in HL, FROM on stack
;V: AF, BC, DE, HL, IX, stack
;R: modified SET on stack (FROM removed)
RangeToSet:	POP	IX		;return address
		POP	DE		;FROM to DE
		LD	A,L
		SUB	E		;TO smaller than FROM?
		JR	C,ToSet2	;yes -> do nothing
		INC	A		;number of elements, which make
		LD	C,A		;up the subrange to C
		LD	B,E		;FROM to B
		CALL	PtrMskSet	;calculate pointer and mask
		LD	E,A		;mask to E
		LD	B,C		;number of elements to B (counter)
		;The desired bits are first set in A. Only when A is full
		;or at the end the bits are transferred into the SET from A.
		XOR	A		;A is empty
RangeToS1:	OR	E		;set next bit in A
		SLA	E		;shift mask. Is A now full?
		JR	NC,RangeToS2	;no -> continue
		OR	(HL)		;yes, transfer bits
		LD	(HL),A
		INC	HL
		XOR	A		;new byte: A is empty,
		LD	E,1		;start with bit 0
RangeToS2:	DJNZ	RangeToS1	;until desired number of bits is set
		JR	ToSet1		;transfer remaining bits

;calculate pointer and mask
;E: bit number in B, SET on stack (under return address)
;V: AF, B, HL
;R: bit mask in A, pointer to corresponding byte of the SET in HL
PtrMskSet:	LD	A,B
		AND	0F8h		;calculate byte offset
		RRCA
		RRCA
		RRCA
		ADD	A,2		;two more because of return address
		LD	L,A
		LD	H,0
		ADD	HL,SP		;ok, now we have the pointer
		LD	A,B		;this is the bit number
		AND	7		;in the byte
		INC	A
		LD	B,A
		XOR	A
		SCF			;calculate mask
PtrMskS1:	RLA
		DJNZ	PtrMskS1
		RET

;store data (in memory)
;----------------------
;(subroutines are provided only for non-trivial data types)

;store REAL number
;E: destination address in HL, REAL number in BCDEHL'
;V:/R: registers swapped (EXX), old HL' destroyed
StoreReal:	PUSH	HL		;put destination address on stack
		EXX			;get REAL number
		EX	DE,HL		;HL has to be stored first, thatswhy we
		EX	(SP),HL		;put it to DE. Get pointer and put old DE
		LD	(HL),E		;on stack. Now store old HL.
		INC	HL
		LD	(HL),D
		INC	HL
		POP	DE		;get old DE and store it
		LD	(HL),E
		INC	HL
		LD	(HL),D
		INC	HL
		LD	(HL),C		;at last store BC
		INC	HL		;everything Lowbyte first, of course
		LD	(HL),B
		RET

;store string
;E: string on stack (under return address, first byte is length),
;   destination address in HL, available destination space in B
;V: AF, BC, DE, HL, IX, stack
;R: pointer behind stored string in DE
StoreString:	POP	IX		;save return address
		LD	A,B		;save available space
		EX	DE,HL		;destination address to DE
		LD	HL,0
		LD	B,H		;B := 0
		ADD	HL,SP		;calculate pointer to beginning of string
		LD	C,(HL)		;get string length
		PUSH	HL
		ADD	HL,BC		;calculate pointer to end of string
StoreString1:	INC	HL		;and save it on stack
		EX	(SP),HL		;get back pointer to beginning
		CP	C		;look if space is sufficient
		JR	C,StoreString2	;no -> take max. allowable length
		LD	A,C		;yes, take string length
StoreString2:	LD	(DE),A		;store length now
		INC	DE
		INC	HL
		OR	A		;if the string is non-empty
		JR	Z,StoreString3	;then copy it
		LD	C,A
		LDIR
StoreString3:	POP	HL		;get saved end of string
		LD	SP,HL		;becomes new SP
		JP	(IX)		;return

;store string
;E: string on stack (under return address, first byte is length),
;   destination address on stack (under string),
;   available destination space in B
;V: AF, BC, DE, HL, IX, stack / R:-
StoreStringS:	POP	IX		;save return address
		LD	A,B
		LD	HL,0
		LD	B,H
		ADD	HL,SP
		LD	C,(HL)
		PUSH	HL
		ADD	HL,BC		;calculate pointer to end of string
		INC	HL		;get destination address from behind end of string
		LD	E,(HL)
		INC	HL
		LD	D,(HL)
		JR	StoreString1	;continue as before

;store SET (partially)
;E: SET on stack, destination address in HL,
;   B = first byte to be copied
;   C = number of bytes to be copied
;V: F, BC, DE, HL, stack / R:-
StoreSet:	POP	IX
		EX	DE,HL		;destination address to DE
		LD	L,B		;calculate pointer to first
		LD	H,0		;byte to be copied
		LD	B,H		;B := 0
		ADD	HL,SP
		LDIR			;copy
		LD	HL,32
StoreSet1:	ADD	HL,SP		;delete SET from stack
		LD	SP,HL
		JP	(IX)

;store SET (partially)
;E: SET on stack, destination address on stack (under SET),
;   B = first byte to be copied
;   C = number of bytes to be copied
;V: F, BC, DE, HL, stack / R:-
StoreSetS:	POP	IX
		LD	HL,32		;calculate pointer behind SET
		ADD	HL,SP
		LD	E,(HL)		;get destination address from there
		INC	HL
		LD	D,(HL)
		LD	L,B		;same as above
		LD	H,0
		LD	B,H
		ADD	HL,SP
		LDIR
		LD	HL,34		;but now delete destination
		JR	StoreSet1	;address from stack too

;get ARRAY OF CHAR as string (copy to stack)
;E: address (source) in HL, length in B
;V: AF, BC, DE, HL, IX, stack / R:-
LoadCharArr:	POP	IX
		EX	DE,HL		;source address to DE
		LD	A,B		;reserve B byte space on stack
		CPL
		LD	L,A
		LD	H,0FFh
		ADD	HL,SP
		LD	SP,HL
		LD	(HL),B		;store length
		INC	HL
		LD	C,B
		LD	B,0
		EX	DE,HL
		LDIR			;copy string
		JP	(IX)

;range checks
;------------

;index check
;E: DE, HL
;V: F
;R: NZero, Carry if DE > HL
;   else no return
IndexCheck:	OR	A
		SBC	HL,DE
		ADD	HL,DE
		RET	C
		LD	A,90h		;Index out of range
IndexChk1:	JP	RunTimErrX

;subrange check
;E: BC, DE, HL
;V: F
;R: NZero, Carry if BC+DE > HL
;   else no return
SubrangeCheck:	OR	A
		SBC	HL,DE
		OR	A
		SBC	HL,BC
		JR	NC,SubrangeCh1
		ADD	HL,BC
		ADD	HL,DE
		RET

SubrangeCh1:	LD	A,91h		;Scalar or subrange out of range
IF (TPVers1 OR TPVers2)
		JR	IndexChk1
ELSE	;TPVers3
		JP	RunTimErrX
ENDIF

;support routines for FOR loop
;-----------------------------

;FOR xx := from TO to
;E: HL = to, DE = from (both integers)
;V: F, DE, HL
;R: HL = from, DE = loops needed (0 if from >= to)
ForTo:		OR	A		;clear Carry
		SBC	HL,DE		;compare borders
		EX	DE,HL
ForTo1:		INC	DE		;Overflow?
		JP	PE,ForTo2	;yes -> then swap borders
		RET	P
		JR	ForTo3

ForTo2:		RET	M
ForTo3:		LD	DE,0		;reject loop
		RET

;FOR xx := from DOWNTO to
;E: HL = to, DE = from (both integers)
;V: F, DE, HL
;R: HL = from, DE = loops needed (0 if to >= from)
ForDownTo:	PUSH	DE		;save from
		EX	DE,HL		;swap boarders
		OR	A
		SBC	HL,DE
		EX	DE,HL
		POP	HL		;from
		JR	ForTo1

;comparisons
;-----------
;Data types INTEGER, BYTE, CHAR and BOOLEAN are treated
;in the same manner (only loading/storing is different).

;X = Y Integer
;E: X in DE, Y in HL
;V: F, HL
;R: HL=1 (TRUE) if X = Y, else HL=0 (FALSE)
EQInt:		OR	A
		SBC	HL,DE
Equal:		LD	HL,1
		RET	Z
		DEC	HL
		RET

;X = Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, HL / R: as in EQInt
EQReal:		CALL	CompReals
		JR	Equal

;X = Y String
;E: X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IY, SP / R: as in EQInt
EQString:	CALL	CompStrings
		JR	Equal

;X <> Y Integer
;E: X in DE, Y in HL
;V: F, HL
;R: HL=1 (TRUE) if X <> Y, else HL=0 (FALSE)
NEInt:		OR	A		;clear Carry
		SBC	HL,DE
NotEqual:	LD	HL,1
		RET	NZ
		DEC	HL
		RET

;X <> Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, HL / R: as in NEInt
NEReal:		CALL	CompReals
		JR	NotEqual

;X <> Y String
;E: X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IY, SP / R: as in NEInt
NEString:	CALL	CompStrings
		JR	NotEqual

;X >= Y Integer
;E: X in DE, Y in HL
;V: AF, DE, HL
;R: HL=1 (TRUE) if X >= Y, else HL=0 (FALSE)
GEInt:		CALL	CompInts
GreaterOrEqu:	LD	HL,1
		RET	NC
		DEC	HL
		RET

;X >= Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, HL / R: as in GEInt
GEReal:		CALL	CompReals
		JR	GreaterOrEqu

;X >= Y String
;E: X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IY, SP / R: as in GEInt
GEString:	CALL	CompStrings
		JR	GreaterOrEqu

;X <= Y Integer
;E: X in DE, Y in HL
;V: AF, DE, HL
;R: HL=1 (TRUE) if X <= Y, else HL=0 (FALSE)
LEInt:		CALL	CompInts
LessOrEqual:	LD	HL,1
		RET	Z
		RET	C
		DEC	HL
		RET

;X <= Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, HL / R: as in LEInt
LEReal:		CALL	CompReals
		JR	LessOrEqual

;X <= Y String
;E: X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IY, SP / R: as in LEInt
LEString:	CALL	CompStrings
		JR	LessOrEqual

;X > Y Integer
;E: X in DE, Y in HL
;V: AF, DE, HL
;R: HL=1 (TRUE) if X > Y, else HL=0 (FALSE)
GTInt:		CALL	CompInts
Greater:	LD	HL,0
		RET	Z
		RET	C
		INC	HL
		RET

;X > Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, HL / R: as in GTInt
GTReal:		CALL	CompReals
		JR	Greater

;X > Y String
;E: X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IY, SP / R: as in GTInt
GTString:	CALL	CompStrings
		JR	Greater

;X < Y Integer
;E: X in DE, Y in HL
;V: AF, DE, HL
;R: HL=1 (TRUE) if X < Y, else HL=0 (FALSE)
LSInt:		CALL	CompInts
Less:		LD	HL,1
		RET	C
		DEC	HL
		RET

;X < Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, HL / R: as in LSInt
LSReal:		CALL	CompReals
		JR	Less

;X < Y String
;E: X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IY, SP / R: as in LSInt
LSString:	CALL	CompStrings
		JR	Less

;INTEGER arithmetic
;------------------

.COMMENT %----------------------------------------------------------------
Basics:

An INTEGER number is 2 byte (16 bit) long.
The most significant bit tells about the polarity of the number.
Negative numbers are represented as 2s complement of their positive
counterparts. Numbers in the range of -32768 to 32767 can be represented.

INTEGER numbers are usually kept in registers HL and DE and are stored
in memory in Intel byte order, i.e. LowByte first. Most simple INTEGER-
operations are performed ad hoc.
----------------------------------------------------------------% ;COMMENT

;sqr(X)
;Integer square
;E: X in HL
;V: AF, BC, DE, HL
;R: X^2 in HL
SqrInt:		LD	D,H		;X^2 is implemented
		LD	E,L		;as X * X
;<fall through>

;X * Y
;E: X in DE, Y in HL
;V: AF, BC, DE, HL
;R: X * Y in HL
MultInt:	LD	C,E		;free up HL
		LD	B,D
		EX	DE,HL		;Y now in DE, X in BC
		LD	HL,0		;initialise 16 bit accumulator
		LD	A,D
		OR	A		;is a 16 bit multipl. necessary?
		LD	A,16		;yes -> then we need 16 loops
		JR	NZ,MultInt1
		LD	D,E		;no, then shift left 8 times
		LD	A,8		;and do only 8 loops
MultInt1:	ADD	HL,HL		;shift left accumulator and Y
		EX	DE,HL
		ADD	HL,HL
		EX	DE,HL
		JR	NC,MultInt2	;if a bit fell out
		ADD	HL,BC		;then add X to accu
MultInt2:	DEC	A
		JR	NZ,MultInt1
		RET

;X div Y
;E: X in DE, Y in HL
;V: AF, BC, DE, HL
;R: no return if Y=0
;   else X div Y in HL, X mod Y in DE, abs(Y) in BC
DivInt:
IF TPVers3
		LD	A,H		;Y=0?
		OR	L
		JP	Z,DivByZero	;yes -> Division by zero
ENDIF
		LD	A,H		;compare sign of X and Y
		XOR	D		;and save result
		PUSH	AF
		CALL	AbsoluteHL	;from now on work with
		EX	DE,HL		;positive numbers only
		CALL	AbsoluteHL
		EX	DE,HL
		LD	B,H
		LD	C,L
		XOR	A		;clears Carry
		LD	H,A
		LD	L,A
		LD	A,17
	;HL together with DE form a 32 bit shift register.
	;Every cycle HL receives one bit from X and DE
	;receives one bit result.
DivInt1:	ADC	HL,HL
		SBC	HL,BC		;if Y fits into it, a
		JR	NC,DivInt2	;1 is shifted to the result
		ADD	HL,BC		;else a 0.
		SCF
DivInt2:	CCF
		RL	E
		RL	D
		DEC	A
		JR	NZ,DivInt1
		EX	DE,HL		;X div Y in HL, and X mod Y in DE
		POP	AF
	;Now care for signs: if X and Y had the same sign we are finished.
	;Else we have to flip the sign of the result.
		RET	P
		JR	NegateHL

;random(X)
;Return random number (Integer)
;E: X in HL
;V: AF, BC, DE, HL
;R: no return if X=0
;   else number between 0 and X-1 in HL
RandInt:	PUSH	HL		;save X
		CALL	Random		;kick random number generator
		SRL	H		;limit to positive numbers
		RR	L
		POP	DE
		EX	DE,HL		;calculate mod X
;<fall through>

;X mod Y
;E: X in DE, Y in HL
;V: AF, BC, DE, HL
;R: no return if Y=0
;   else X mod Y in HL
ModInt:		CALL	DivInt		;calcuate X mod Y
		EX	DE,HL
		BIT	7,D		;if X div Y negative, then
		RET	Z		;X mod Y negative too
		JR	NegateHL

.COMMENT %----------------------------------------------------------------
recommended patch in c't 10/86 p.84f >>
V2.00:			V3.00:
070Fh	CDh -> D5h	0745h	CDh -> D5h
0710h	DEh -> CDh	0746h	0Fh -> CDh
0711h	06h -> DEh	0747h	07h -> 0Fh
0712h	EBh -> 06h	0748h	EBh -> 07h
0713h	CBh -> F1h	0749h	CBh -> F1h
0714h	7Ah -> EBh	074Ah	7Ah -> EBh
0715h	C8h -> F0h	074Bh	C8h -> F0h

ss code:	PUSH	DE		;save X
		CALL	DivInt		;calculate X mod Y
		POP	AF
		EX	DE,HL		;if X negative, then
		RET	P		;X mod Y negative too
		JR	NegateHL

What results do we expect from a DIV b and a MOD b?
With a and b positive, there is no doubt that DIV returns
the integer part of the division result while MOD returns
the rest of the division, e.g.:
17 DIV 6 = 2, 17 MOD 6 = 5.

The following holds:
1) 0 <= a MOD b < b
2) a = (a DIV b) * b + (a MOD b)

With a negative:
Both versions (original and patched) mirror at zero.
-17 DIV 6 = -2 seems logical at first glance. But then the result of MOD
is negative! More correct would be not to mirror at zero but to continue
the same procedure below zero (imagine the numbers written on an axis).
Then we would have -17 DIV 6 = -3 and -17 MOD 6 = 1!

With b negative:
One cannot imagine things so easily with the axis. And theorem (1)
can't be fulfilled. And this is where there is a difference
between the original and the patched version: the original version
doesn't fulfill theorem (2) while the patched one does.

Sign:
 a b ! a div b ! a mod b
-----+---------+---------
 + + !    +    !    +
 - + !    -    !    -	 v--- patched
 + - !    -    !    -    +
 - - !    +    !    +    -
		    ^-------- original
----------------------------------------------------------------% ;COMMENT

;X shl Y
;Shift left
;E: X in DE, Y in HL
;V: AF, B, DE, HL
;R: (Y in DE)
; if shift count out of range:
;   HL = A =0, B unchanged
; else:
;   HL = X shl Y, A=Y, B=0
ShLeft:		CALL	Range16
		RET	Z
ShLeft1:	ADD	HL,HL
		DJNZ	ShLeft1
		RET

;X shr Y
;Shift right
;E: X in DE, Y in HL
;V: AF, B, DE, HL
;R: (X in DE)
; if shift count out of range:
;   HL = A =0, B unchanged
; else:
;   HL = X shr Y, A=Y, B=0
ShRight:	CALL	Range16
		RET	Z
ShRight1:	SRL	H
		RR	L
		DJNZ	ShRight1
		RET

;Range check 0-15
;E: shift count in HL, value in DE
;V: AF, B, DE, HL
;R: shift count in DE, NCarry
;   value in HL, A=B= shift count, N/Zero if shift count in 0-15
;   else A=HL=0, Zero, B unchanged
Range16:	EX	DE,HL
		LD	A,D
		OR	A
		JR	NZ,ClearHL
		LD	A,E
		CP	16
		JR	NC,ClearHL
		LD	B,A
		OR	A
		RET

;HL := 0
;E:- / V: AF, HL / R: Zero, A=0, HL=0
ClearHL:	XOR	A
		LD	H,A
		LD	L,A
		RET

;compare two integers (signed!)
;E: HL, DE
;V: AF (DE and HL are swapped)
;R:  Carry	  if HL > DE
;   NCarry,  Zero if HL = DE
;   NCarry, NZero if HL < DE
CompInts:	EX	DE,HL
		LD	A,H
		XOR	D		;minus, if HL and DE have
		LD	A,H		;different sign
		JP	M,CompInts1
		CP	D		;same sign
		RET	NZ		;compare Highbytes first
		LD	A,L		;if equal, compare
		CP	E		;Lowbytes also
		RET

;Signs of HL and DE differ. Whether HL or DE is bigger depends
;on the sign of HL (and DE).
CompInts1:	RLA			;sign of HL (former DE) -> Carry
		RET

;abs(X) Integer
;calculate absolute value
;E: X in HL / V: AF (without Carry), HL / R: HL
AbsoluteHL:	BIT	7,H		;do nothing if positive
		RET	Z
;<fall through>

;calculate 2s complement (HL := 0-HL, "negate")
;E: HL / V: AF (only flags H,N), HL / R: HL
NegateHL:	LD	A,H		;calculate 1s complement
		CPL			;first
		LD	H,A
		LD	A,L
		CPL
		LD	L,A
		INC	HL		;then +1: gives 2s complement
		RET

;odd(X)
;HL := HL AND $0001
;E: X in HL / V: AF, HL / R: odd(X) in HL
OddInt:		LD	A,L
		AND	1
		LD	L,A
		LD	H,0
		RET

;random number generator
;E:- V: AF, BC, DE, HL
;R: 32 bit random number in BCDE (HighWord also in HL)
;Literature:
;Donald E. Knuth, The Art of Computer Programming, Vol 2
;(Seminumerical Algorithms), Addison-Wesley Publishing Company
;Method:
;Linear congruence method: Xn+1 := (a * Xn + c) mod m
;hier:	Seed := (Seed + (Seed shl 7) + 361962E9h) mod 2^32
;also:	Seed := ((Seed * 129) + 907,633,385d) mod 2^32
;Demands:
;(let b = a-1, k the size of the words of the computer in bits und w = 2^k)
;k = 32, m = w = 2^32
;a = 129, b = 128 = 2^7
;c = 907633385 = 5 * 107 * 1696511
;# m should be as big as possible so that the random number sequence
;  doesn't repeat too fast -> is big enough here
;# if m is a power of 2 the bits on the right side of the produced random
;  numbers are not that much random, m should therefore not be w, but
;  w-1 or w+1 -> was not considered here
;  Knuth: "In most applications, the low-order bits are insignificant, and
;  the choice m = w is quite satisfactory - provided that the programmer
;  using the random numbers does so wisely."
;# The method should produce exactly m different numbers before repeating.
;  This can be done if and only if
;  i)   c is relatively prime to m
;  here: m is a power of 2, c can't be divided by 2 -> ok
;  ii)  b is a multiple of p for all prime numbers p that divide m
;  here: there is only one prime number that divides m, the 2, and b
;	 is a power of 2 -> ok
;  iii) b is a multiple of 4, if m is a multiple of 4
;  here: m is multiple of 4, b too -> ok
;# potency of a linear congruence sequence with maximum period:
;  smallest integer number s with b^s = 0 mod m
;  here: s = 5, because (2^7)^5 = 2^35 = 0 mod 2^32
;  Knuth: "A potency of at least 5 would seem to be required for
;  sufficiently random values" -> ok
;# more informations about the quality of this algorithm can be
;  obtained with the spectral test (see Knuth), which I don't
;  fully understand and therefore omit
;# I would really like to know how to construct a good random number
;  generator. Finding one by trial and error is not elegant and even
;  with the help of a computer very time consuming ...
Random:		LD	BC,(Seed+2)	;get seed
		LD	DE,(Seed)
		PUSH	BC		;save it
		PUSH	DE
		LD	A,B		;shift left 8 times
		LD	B,C
		LD	C,D
		LD	D,E
		LD	E,0
		RRA			;and back one time
		RR	B
		RR	C
		RR	D
		RR	E
		POP	HL
		ADD	HL,DE		;add to saved seed
		EX	DE,HL
		POP	HL
		ADC	HL,BC
		LD	B,H
		LD	C,L
		LD	HL,62E9h	;add a constant
		ADD	HL,DE
		LD	(Seed),HL	;gives new seed
		EX	DE,HL
		LD	HL,3619h
		ADC	HL,BC
		LD	(Seed+2),HL
		LD	B,H		;return seed in BCDE
		LD	C,L		;HighWord in HL also
		RET

;Convert unsigned word into Ascii representation (decimal),
;leading zeros are suppressed,
;ciffers are inserted flush-left in the buffer.
;E: Number in HL, pointer to buffer in IX
;V: AF, B, DE, HL, IX
;R: pointer behind last ciffer in IX
ItoA:		LD	B,0		;flag: =0 while there are
		LD	DE,10000	;leading zeros
		CALL	ItoA1		;divide and insert result
		LD	DE,1000
		CALL	ItoA1
		LD	DE,100
		CALL	ItoA1
		LD	E,10
		CALL	ItoA1		;last ciffer:
		LD	A,L		;do not divide, issue
		JR	ItoA3		;everytime even if zero

ItoA1:		XOR	A		;counter
		;subtract DE of HL until one time too much
ItoA2:		INC	A
		SBC	HL,DE
		JR	NC,ItoA2
		;correct the "too much"
		ADD	HL,DE		;suppose it's not zero,
		INC	B		;keeping old value of flag
		DEC	A		;correct and look if zero
		JR	NZ,ItoA3	;no -> flag already set
		;we have a zero, so restore flag to its previous value
		;if flag was reset before, then we have a leading zero
		DEC	B
		RET	Z		;flag=0? Yes -> suppress
ItoA3:		ADD	A,'0'		;convert to ASCII character
		LD	(IX),A		;and store into buffer
		INC	IX
		RET

;AtoI
;Convert number (in ASCII representation) into integer.
;Representation is decimal or, with a leading '$', hexadecimal.
;E: pointer to string in IX
;V: AF, C, DE, HL, IX
;R: pointer to first non-ciffer in IX
;   Carry if number out of range
;   else number in HL
AtoI:		LD	A,(IX)		;first character
		SUB	'$'		;a dollar sign?
		LD	C,A		;save as flag
		LD	HL,0
		JR	NZ,AtoI2
AtoI1:		INC	IX
AtoI2:		LD	A,(IX)		;get next character
		CALL	ToUpper		;in case we have a hexadecimal number
		SUB	'0'		;smaller than '0'?
		JR	C,AtoI5		;yes -> exit
		CP	10		;smaller than 10?
		JR	C,AtoI3		;yes -> is always ok
		INC	C
		DEC	C		;are we reading a hexadecimal number?
		JR	NZ,AtoI5	;no -> exit
		SUB	7
		CP	10		;is it between '9' and 'A'?
		JR	C,AtoI5		;yes -> exit
		CP	16		;greater than 'F'?
		JR	NC,AtoI5	;yes -> exit
AtoI3:	;now valid (hex-)ciffer in A
		LD	D,H		;save times 1
		LD	E,L
		ADD	HL,HL
		RET	C
		ADD	HL,HL		;calculate times 4
		RET	C
		INC	C
		DEC	C		;are we reading a hexadecimal number?
		JR	NZ,AtoI4	;no -> keep times 1
		LD	D,H		;yes, save times 4 instead
		LD	E,L
AtoI4:		ADD	HL,DE
		RET	C
		ADD	HL,HL		;times 10 or 16 respectively
		RET	C
		LD	E,A
		LD	D,0		;add ciffer
		ADD	HL,DE
		RET	C
		JR	AtoI1		;continue until error

AtoI5:	;illegal character encountered
		LD	A,C		;hexadecimal number?
		OR	A
		RET	Z		;yes -> 0-0FFFFh legal
		LD	A,H		;decimal: only 0-32767 legal!
		ADD	A,A		;returns Carry if more
		RET

;END OF TP-RTL2.INC
;------------------
