;--------------------------------
;TP-RTL3.INC	ho	01-MAR-94
;--------------------------------
;V1.00: 07CAh-0A80h
;V2.00: 0807h-0ABDh	V2.01: 0801h-0AB7h
;V3.00: 083Dh-0AF4h	V3.01: 083Eh-0AF5h

		IDENT	TP-RTL3.INC

;STRING arithmetic
;-----------------

.COMMENT %----------------------------------------------------------------
Basics:

In Turbo Pascal a STRING with maximum length n can be accessed like
an ARRAY[0..n] OF CHAR. The CHAR with index 0 is in fact a BYTE that
holds the actual length of the STRING, which can be anything between
0 and 255.
----------------------------------------------------------------% ;COMMENT

;concat(X, Y) or X + Y String
;concat with more than 2 arguments is converted into a series of
;concat subroutine calls by the compiler.
;E: strings X and Y on stack (Attention: X is under Y!)
;V: AF, BC, DE, HL, IX, stack
;R: no return if total length > 255
;   else string XY on stack
Concat:		POP	IX
		POP	HL		;get length of string Y
		PUSH	HL
		LD	A,L
		LD	H,0
		INC	HL
		ADD	HL,SP		;calculate pointer to X
		LD	C,(HL)		;get length of X and calculate
		ADD	A,C		;total length. Greater than 255?
		JR	C,StrLenError	;yes -> error
		LD	(HL),A		;save total length in length byte of X
		EX	DE,HL		;save pointer to X
		LD	HL,0
		LD	B,H
		SBC	HL,BC		;reserve space for X on stack
		ADD	HL,SP		;(in front of Y)
		LD	SP,HL
		EX	DE,HL
		PUSH	HL		;one more for (total) length byte
		INC	BC		;copy X in front of Y (overwriting
		LDIR			;Y's length byte)
		EX	DE,HL		;DE now points behind old X
		POP	HL		;HL points to old X
		DEC	HL		;point HL to end of XY,
		DEC	DE		;point DE to end of old X
		LD	C,A		;length of XY
		INC	BC		;one more for length byte
		LDDR			;copy over old X (move up)
		EX	DE,HL		;HL points in front of string XY
		INC	HL		;now to XY
		LD	SP,HL		;this is our new SP
		JP	(IX)

StrLenError:	LD	A,10h		;String length error
		JP	RunTimErr

;copy(X, V, W)
;take a part of string X, starting with index V, length W
;E: W in HL, string X and index V on stack
;V: AF, BC, DE, HL, IX, stack
;R: no return if index V out of range
;   else part of string on stack
Copy:		POP	IX
		CALL	Range256	;W: force legal length
		LD	D,A		;save
		POP	HL		;get index
		CALL	ChkStrInx	;check it
		LD	E,A		;and save it
		POP	HL		;get length of X
		PUSH	HL
		LD	A,L
		SUB	E		;index > length?
		JR	C,Copy1		;yes -> return empty string
		INC	D		;W =0?
		DEC	D
		JR	Z,Copy1		;yes -> return empty string
		CP	D		;V + W > length?
		JR	C,Copy2		;yes -> return tail
	;create partial string (by copying to end of string)
		LD	C,D		;W is new length
		LD	B,0
		LD	H,B
		ADD	HL,SP		;calculate pointer to end of X
		LD	A,E
		ADD	A,D		;calculate pointer to last character
		LD	D,H		;of the partial string
		LD	E,L
		DEC	A
		LD	L,A
		LD	H,B
		ADD	HL,SP
		LD	A,C
		LDDR			;copy partial string to end of string
		EX	DE,HL		;pointer to beginning of partial string in HL
		JR	Copy4

Copy1:	;return empty string, i.e. tail with length 0
		XOR	A		;L still holds length of X
		JR	Copy3

Copy2:	;return tail (cut away head of string)
		INC	A
		LD	L,E
		DEC	L		;calculate pointer to length byte
Copy3:		LD	H,0		;of tail
		ADD	HL,SP
Copy4:		LD	(HL),A		;insert length
		LD	SP,HL		;free up stack space
		JP	(IX)

;length(X)
;E: string on stack
;V: AF, HL, IX, stack
;R: length of X in HL
Length:		POP	IX
		POP	HL		;calculate length of X
		PUSH	HL
		LD	A,L
		LD	H,0
		INC	HL
		ADD	HL,SP		;remove string from stack
		LD	SP,HL
		LD	L,A		;return length
		LD	H,0
		JP	(IX)

;pos(X, Y)
;Search the first occurance of string X in string Y
;E: X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IX, IY, stack
;R: 0 in HL if X longer than Y or not found,
;   else index in HL
Pos:		POP	IX
		LD	HL,0
		LD	D,H
		ADD	HL,SP		;calculate pointer to Y
		LD	E,(HL)		;get length of Y
		LD	C,E		;and save it
		INC	HL
		PUSH	HL
		ADD	HL,DE		;calculate pointer to X
		LD	E,(HL)		;get length of X
		LD	B,E		;and save it
		INC	HL
		PUSH	HL
		ADD	HL,DE
		PUSH	HL
		POP	IY		;= pointer behind both strings
		POP	DE		;= pointer to X
		POP	HL		;= pointer to Y
		LD	A,C
		SUB	B		;X bigger than Y?
		JR	C,Pos3		;yes -> can't find that
		INC	A
		LD	C,A		;the two lengths differ by that much,
Pos1:		PUSH	BC		;i.e. we have to test that often
		PUSH	DE
		PUSH	HL
Pos2:		LD	A,(DE)		;compare characters
		CP	(HL)		;identical?
		JR	Z,Pos4		;yes -> continue comparing
		POP	HL		;no, abort test
		POP	DE
		POP	BC
		INC	HL		;shift X right one character with respect to Y
		DEC	C		;all possibilities tested?
		JR	NZ,Pos1		;no -> try again

Pos3:	;not found
		LD	HL,0		;return zero
		JR	Pos5		;exit

Pos4:	;test next characters
		INC	HL
		INC	DE		;end of string X?
		DJNZ	Pos2		;no -> continue comparing
		POP	DE		;yes, found!
		POP	HL
		POP	BC
		LD	HL,0
		ADD	HL,SP		;calculate pointer to Y
		EX	DE,HL		;and subtract from pointer to occurance
		SBC	HL,DE		;gives index
Pos5:		LD	SP,IY		;remove both strings from stack
		JP	(IX)

;delete(X, V, W)
;Delete W characters in string X starting at index V.
;E: W in HL, pointer to string and index on stack
;V: AF, BC, DE, HL, IX, stack
;R: no return if index V out of range
Delete:		POP	IX
		CALL	Range256	;W: force valid length
		LD	C,A		;save it
		POP	HL		;get index
		CALL	ChkStrInx	;check it
		LD	E,A		;and save it
		POP	HL		;get pointer to string
		LD	A,(HL)		;get length
		SUB	E		;smaller than index?
		JR	C,Delete2	;yes -> do nothing
		INC	C
		DEC	C		;W=0?
		JR	Z,Delete2	;yes -> do nothing
		SUB	C		;index+W > length ?
		JR	C,Delete1	;yes -> only decrease length
		PUSH	AF		;no, save length of rest of string
		LD	A,(HL)		;get length
		SUB	C		;decrease
		LD	(HL),A
		LD	B,0
		LD	D,B		;point to first
		ADD	HL,DE		;character to be deleted
		LD	D,H
		LD	E,L		;point to first character
		ADD	HL,BC		;of rest of string
		POP	AF
		INC	A		;get length of rest of string
		LD	C,A
		LDIR			;copy downward over the part
		JR	Delete2		;that is to be deleted

Delete1:	DEC	E		;rest of string is to be deleted
		LD	(HL),E		;so index is new length
Delete2:	JP	(IX)

;insert(X, Y, V)
;Insert string X into string Y at position V
;E: max. length of string Y in B, index V in HL,
;   pointer to string Y and to string X on stack
;V: AF, BC, DE, HL, IX, stack
;R: no return if index V out of range
Insert:		POP	IX
		CALL	ChkStrInx	;check index
		LD	C,A		;and save it
		POP	DE		;pointer to string Y to DE
		LD	(InsertPtr),DE	;save it
		LD	HL,0
		ADD	HL,SP		;pointer to String X to HL
		LD	A,(DE)		;current length of Y
		PUSH	AF		;save it
		ADD	A,(HL)		;add length of X. Bigger than 255?
		JR	C,Insert1	;yes -> take max. length
		CP	B		;bigger than max. length?
		JR	C,Insert2	;no -> take this length
Insert1:	LD	A,B		;take max. length
Insert2:	LD	(DE),A		;insert new length
		POP	AF		;old length of Y
		LD	D,A		;save it
		LD	E,(HL)		;save length of X
		SUB	C		;index > old length of Y?
		JR	C,Insert5	;yes -> glue it at the tail
	;Reserve space for X in Y by moving the tail of Y starting with index V
	;At first calulcate the length of the tail of Y (= Y')
	;honoring the maximum length
		INC	A		;calculate the length of Y' without
		LD	L,A		;honoring the max. length
	;If the length of X+Y is smaller than the max. length, this
	;value is used. Else the length of Y' has to be computed
	;based on the max. length
		LD	A,D
		ADD	A,E		;old length of Y + length of X
		JR	C,Insert3	;bigger than 255? yes -> recalculate
		CP	B		;bigger than max. length?
		LD	A,L
		JR	C,Insert4	;no -> calculated value ok
	;Calculate length of Y' based on the maximum length. If X or
	;X+index bigger than max. length we have to move nothing
Insert3:	LD	A,B		;max. length
		SUB	E		;- length of X
		JR	C,Insert6	;X itself too big -> do nothing
		SUB	C		;- index
		JR	C,Insert6	;X+index too big -> do nothing
		INC	A
	;now length in A
Insert4:	OR	A		;length =0?
		JR	Z,Insert6	;then do nothing too
	;Move
		PUSH	BC		;save length informations
		PUSH	DE
	;source = pointer to last character in Y, that is to be kept
		LD	HL,(InsertPtr)	;get pointer to Y
		LD	E,A		;extend length of Y' and
		DEC	E		;index to a word
		LD	D,0
		LD	B,D
		ADD	HL,DE		;+ length of Y'
		ADD	HL,BC		;+ index = source
		POP	DE		;get length of X
		PUSH	DE		;and save it again
		PUSH	HL		;save source
	;destination = source + length of X
		LD	D,B		;extend length of X to a word
		ADD	HL,DE
		EX	DE,HL		;destination to DE
		POP	HL		;restore source
		LD	C,A		;length of Y'
		LDDR			;copy
		POP	DE		;restore length informations
		POP	BC
		JR	Insert6

	;glue X to tail of Y
Insert5:	LD	A,D		;old length of Y
		INC	A		;=255?
		JR	Z,Insert8	;yes -> no space for X
		LD	C,A		;index := length of Y + 1

	;glue X to tail of Y / copy X into Y
	;calculate length of part of X, that is to be copied (=X')
Insert6:	LD	A,B
		SUB	C		;max. length of X :=
		INC	A		; max. length - index + 1
		CP	E		;length of X bigger?
		JR	C,Insert7	;yes -> take max. length of X
		LD	A,E		;no, take length of X
Insert7:	OR	A		;length =0?
		JR	Z,Insert8	;yes -> don't copy
	;copy
		LD	HL,(InsertPtr)	;pointer to Y
		LD	B,0		;extend index to a word
		ADD	HL,BC		;destination := Y[index]
		EX	DE,HL		;to DE
		LD	HL,1		;don't include length of X
		ADD	HL,SP		;= pointer to X = source
		LD	C,A		;length of X'
		LDIR			;copy
	;finished, remove X from stack
Insert8:	LD	HL,0
		LD	D,H
		ADD	HL,SP		;pointer to X
		LD	E,(HL)		;length of X
		INC	DE		;+1 because of length byte
		ADD	HL,DE		;pointer behind X
		LD	SP,HL		;becomes new SP
		JP	(IX)

;convert string of length 1 to a CHAR
;E: string on stack
;V: F, HL, IX
;R: no return if string not exactly one char long
;   else character in HL
StrToChar:	POP	IX
		POP	HL		;get length + first character
		DEC	L		;length =1?
		JP	NZ,StrLenError	;no -> String length error
		LD	L,H		;yes, return string
		LD	H,0
		JP	(IX)

;convert CHAR to a string of length 1 (used for string + CHAR)
;E: string on stack (under return address), character on stack under it (word wide)
;V: AF, DE, HL
;R: string on stack (under ret.addr.), string of length 1 under it
CharToStr:	LD	HL,2		;because of return address
		LD	D,H		;=0
		ADD	HL,SP		;calculate pointer to string
		LD	E,(HL)		;get length
		INC	DE
		ADD	HL,DE		;pointer behind string (to character)
		LD	A,(HL)		;get character
		LD	(HL),1		;insert length
		INC	HL		;and write character in the
		LD	(HL),A		;following (empty) byte
		RET

;string comparison
;E: strings X, Y on stack (X under Y)
;V: AF, BC, DE, HL, IY, stack
;R: NZero, Carry  if X < Y
;   NCarry, Zero, if X = Y
;   NZero, NCarry if X > Y
CompStrings:	LD	HL,4		;two return addresses
		LD	D,H
		ADD	HL,SP		;calculate pointer to Y
		LD	E,(HL)		;length of Y
		LD	C,E		;save it
		INC	HL		;advance to first char
		PUSH	HL		;and save it
		ADD	HL,DE		;calculate pointer to X
		LD	E,(HL)		;length of X
		LD	B,E		;save it
		INC	HL		;advance to first char
		PUSH	HL
		ADD	HL,DE
		PUSH	HL
		POP	IY		;pointer behind X
		POP	DE		;pointer to first char of X
		POP	HL		;pointer to first char of Y
CompStr1:	;while no difference between X and Y
		XOR	A
		CP	B		;X ends?
		JR	Z,CompStr2	;yes -> look if Y longer
		CP	C		;Y end too?
		JR	NZ,CompStr4	;no -> compare character
		LD	A,B
CompStr2:	;If a string ends, then that one is "bigger" that is longer.
		;If both end, they are same.
		CP	C
CompStr3:	POP	HL		;save the two return
		POP	DE		;addresses
		LD	SP,IY		;remove the two strings
		PUSH	DE		;restore the first return address
		JP	(HL)		;use the other directly

CompStr4:	;compare one character
		LD	A,(DE)		;character from X
		CP	(HL)		;equal to character from Y?
		JR	NZ,CompStr3	;no -> exit
		INC	HL		;yes, advance pointers
		INC	DE
		DEC	B		;and decrease remaining length
		DEC	C
		JR	CompStr1	;continue searching for a difference

;check string index
;legal: 1-255
ChkStrInx:	LD	A,H
		OR	A		;H has to be 0
		JR	NZ,StrInxError
		LD	A,L
		OR	A		;L must'nt be 0
		RET	NZ
StrInxError:	LD	A,11h		;Invalid string index
		JP	RunTimErr

;REAL arithmetic
;---------------

.COMMENT %----------------------------------------------------------------
Basics:

A REAL number is 6 byte (48 bits) long.
Of that, 5 byte (40 bits) are the mantissa and 1 byte (8 bit) are the
characteristic. Characteristic := exponent + 128 (80h), that spares
us from having to deal with the sign of the exponent in a special way.
The mantissa of a b-system is always between 0.1 and 0.xx..., where
x stands for b-1. In a binary system that means, that the mantissa is
always between 0.1 and 0.111...! You'll notice, that the first binary
digit after the decimal point, i.e. the first bit of the mantissa is
ALWAYS a 1. Now if we know that, we don't have to save that! Instead,
the first bit of the mantissa is used for the sign bit.
The situation with REAL numbers is different to the one with INTEGERs
in that negative values are NOT represented as the complement of their
positive counterparts. Instead, always the absolute value is used,
which is marked negative by setting the sign bit.
Normally, the first action when working with REAL numbers is that the
sign bit is extracted and stored seperately while the first bit of
the mantissa is set to 1 to enable correct mathematical operations.

One notable exception is the zero, because this number can't be
represented with the system shown above. Here, every number that
has a zero characteristic means zero, regardless of the mantissa.
But normally, the mantissa is 5 bytes of zero too.

With this system we are able to represent numbers in the range of (absolute):
(01 00 00 00 00 00 =) 2.9387358771E-39 bis (FF FF FF FF FF 7F =) 1.7014118346E+38
with 11 valid (decimal-)digits.
But the input conversion routines only accept numbers in the range of
1E-38 and 1E+38 (excluded).

A REAL number is usually kept in registers BC, DE and HL. B keeps the
most significant byte and L the least significant byte. The 7th bit of
B thus holds the sign, the rest of B together with CDEH form the mantissa,
and L holds the characteristic.
Two REAL values can be stored in the Z80 registers simultanously, one
value in BCDEHL and the other in the shadow registers BCDEHL'.
In memory, the REAL numbers are stored in usual Intel byte order (LowByte
first), i.e. first register L, then H and so on until register B.

The abbreviation "TdM" that is used in the following means:
for a reference, see the book
Bronstein-Semendjajew, "Taschenbuch der Mathematik", Verlag Harri Deutsch,
Thun und Frankfurt/Main, 22. Auflage 1985.
It's the so-called "bible of the mathematicians".
----------------------------------------------------------------% ;COMMENT

;Basic operations (user-oriented routines) >>

;X + Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: no return if overflow
;   else X+Y in BCDEHL
AddReal:	CALL	@AddReal
RealRet:	RET	NC
		LD	A,01h		;Floating point overflow
RealRet1:	JP	RunTimErrX

;X - Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: no return if overflow
;   else X-Y in BCDEHL
SubReal:	CALL	@SubReal
		JR	RealRet

;sqr(X) Real
;E: X in BCDEHL
;V: AF, BC, DE, HL, AF', BCDEHL'
;R: no return if overflow
;   else X^2 in BCDEHL
SqrReal:	CALL	DupReal
;<fall through>

;X * Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: no return if overflow
;   else X * Y in BCDEHL
MultReal:	CALL	@MultReal
		JR	RealRet

;X / Y Real
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: no return if overflow
;   else X / Y in BCDEHL
DivReal:	EXX
		LD	A,L		;Y=0.0?
		OR	A
		EXX			;if so, then complain!
DivByZero:	LD	A,02h		;Division by zero
IF (TPVers1 OR TPVers2)
		JR	Z,RealRet1
ELSE	;TPVers3
		JP	Z,RunTimErrX
ENDIF
		CALL	@DivReal	;else do division
		JR	RealRet

;Basic operations (for internal use) >>

;X + Y Real (internal)
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: Carry if overflow
;   else X+Y in BCDEHL (Y = BCDEHL' unchanged!)
@AddReal:	EXX
		BIT	7,B		;Y negative?
		EXX
		JP	NZ,SubRealAbs	;yes -> calculate -(-X+ABS(Y))
;<fall through>

;come here if X+(positive Y) or X-(negative Y)
;X + ABS(Y)
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: Carry if overflow
;   else X+ABS(Y) in BCDEHL
AddRealAbs:	;first sort out trivial cases
		EXX
		LD	A,L
		OR	A		;Y =0.0? (and clear Carry)
		EXX
		RET	Z		;yes -> do nothing
		EXX
		PUSH	BC		;save Y
		PUSH	DE
		PUSH	HL
		EXX
		LD	A,L
		OR	A		;X =0.0? (and clear Carry)
		JR	NZ,AddRAbs1	;no -> we have to do something
		EXX
		RES	7,B		;yes, return ABS(Y)
		JR	PopXReal

AddRAbs1:	;prepare for addition: shift mantissae
		PUSH	BC		;save sign of X
		SET	7,B		;first bit of mantissa =1
		XOR	A
		EX	AF,AF'		;set Zero'
		EXX
		SET	7,B		;implicitely: ABS(Y)
		LD	A,L		;charact. of Y
		EXX
		SUB	L		;-charact. of X
		JR	Z,AddRAbs3	;ch. Y = ch. X -> do not shift
		JR	NC,AddRAbs2	;ch. Y > ch. X -> shift in opposite direction
		NEG			;make shift count positive
		EX	AF,AF'		;set flag indicating that the numbers have
		DEC	A		;been swapped. Set NZero'
		EX	AF,AF'
		EXX			;swap numbers
AddRAbs2:	CALL	SRLMantisse	;shift right
		INC	L		;adapt charact.
		DEC	A		;enough shifted?
		JR	NZ,AddRAbs2	;no -> continue
		EX	AF,AF'		;have numbers been swapped?
		JR	Z,AddRAbs3	;no -> do not swap back
		EXX
AddRAbs3:	;ABS(Y) is positive. If X positive, then add absolute
		;values, else swap and subtract absolute values
		POP	AF		;get sign of X
		AND	80h		;was X negative?
		JR	NZ,AddRAbs4	;yes -> then subtract
		;add
		CALL	AddMantisse	;did a bit fell out?
		JR	NC,AddRAbs9	;no -> that's it
		CALL	RRMantisse	;yes, stuff it back in
		OR	A		;clear Carry
		INC	L		;and increase charact. Overflow?
		JR	NZ,AddRAbs9	;no -> that's it
		SCF			;yes, return Carry
		JR	PopXReal

AddRAbs4:	;prepare for subtracting absolute values (Y-X)
		CALL	CompMantisse	;look what is greater
		CCF
		PUSH	AF		;save result (Carry, if X > Y)
		JR	Z,AddRAbs7	;if equal, then trivial
		JR	C,AddRAbs5
		EXX			;if X < Y, then swap

AddRAbs5:	;subtract
		CALL	SubMantisse	;subtract the smaller number from the bigger one
AddRAbs6:	BIT	7,B		;norming necessary?
		JR	NZ,AddRAbs8	;no -> maybe need to swap back
		CALL	SLAMantisse	;shift left
		DEC	L		;decrease charact. Number too small?
		JR	NZ,AddRAbs6	;no -> continue norming
AddRAbs7:	CALL	LoadZero	;yes, return zero
AddRAbs8:	POP	AF		;was X > Y?
		;yes, then we have calculated X-Y instead of Y-X
		;and have to keep the result negative

AddRAbs9:	;Carry -> result negative. NCarry -> result positive
		JR	C,AddRAbs10
		RES	7,B
AddRAbs10:	OR	A		;clear Carry
PopXReal:	EXX			;get saved Y
		POP	HL
		POP	DE
		POP	BC
		EXX
		RET

;X - Y Real (internal)
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: Carry if overflow
;   else X-Y in BCDEHL (Y = BCDEHL' unchanged!)
@SubReal:	EXX
		BIT	7,B		;Y negative?
		EXX
		JP	NZ,AddRealAbs	;yes -> calculate X + ABS(Y)

;come here if X+(negative Y) or X-(positive Y)
;- (-X + ABS(Y))
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: Carry if overflow
;   else -(-X+ABS(Y)) in BCDEHL
SubRealAbs:	CALL	NegateReal
		CALL	AddRealAbs
		RET	C
;<fall through>

;negate Real number (flip sign bit)
;E: X in BCDEHL
;V: AF, B
;R: -X in BCDEHL
NegateReal:	INC	L		;is it 0.0?
		DEC	L
		RET	Z		;yes -> do nothing
		LD	A,B		;else flip
		XOR	80h		;sign bit
		LD	B,A
		RET

;X * Y Real (internal)
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: Carry if overflow
;   else X * Y in BCDEHL (Y = BCDEHL' unchanged!)
@MultReal:	EXX
		LD	A,L		;Y = 0.0?
		OR	A
		EXX
		JP	Z,LoadZero	;yes -> then result := 0.0
		LD	A,L		;X = 0.0?
		OR	A
		RET	Z		;yes -> do nothing
		EXX
		ADD	A,L		;add charact.
		EXX
		CALL	MulDivVor
		PUSH	BC		;put X on stack
		PUSH	DE
		PUSH	HL
		ADD	IX,SP		;copy SP to IX
;now multiply.
;Method: comparable to integer multiplication!
		CALL	LoadZero	;RealAccu := 0.0
		EXX
		LD	L,5		;byte counter in char. byte of Y
		EXX			;(cause charact. is already evaluated)
MultReal1:	LD	A,8		;bit counter
		INC	IX		;next byte
		LD	L,(IX)		;get byte (into char. byte of X, see above)
MultReal2:	EX	AF,AF'		;save counter A
		RR	L
		JR	NC,MultReal3	;rotate mantissa of X. If Carry, then
		CALL	AddMantisse	;RealAccu := RealAccu + mantissa of Y
MultReal3:	CALL	RRMantisse	;rotate RealAccu
		EX	AF,AF'		;get bit counter. If the last time:
					;save additional result bit in Carry!
		DEC	A		;finished?
		JR	NZ,MultReal2	;no -> next bit
		EXX
		DEC	L		;decrease byte counter. Finished?
		EXX
		JR	NZ,MultReal1	;no -> next byte
;mantissa of RealAccus is ok now
		LD	L,(IX-5)	;get charact. of result
;norm result:
;if the most significant bit of the mantissa =0, then shift left
;and decrease characteristic (if still >0)
		BIT	7,B		;most significant bit of mantissa =0?
		JR	NZ,MultReal4	;no -> finished with norming
		EX	AF,AF'		;get additional result bit
		CALL	RLMantisse	;and rotate it in
		INC	L		;charact. =0?
		DEC	L
		JR	Z,MultReal4	;yes -> underflow
		DEC	L
MultReal4:	POP	AF		;throw away saved X
		POP	AF
		POP	AF
MulDiv1:	OR	A		;NCarry (ok.)
MulDiv2:	EX	AF,AF'		;save error flag
		POP	AF		;get sign of result
		EXX
		POP	BC		;restore Y
		POP	HL
		EXX
		POP	IX		;restore IX
		RES	7,B		;free up most significant
		OR	B		;bit of the mantissa and
		LD	B,A		;copy in sign
		INC	L
		DEC	L		;result very small?
		CALL	Z,LoadZero	;yes -> set to 0.0
		EX	AF,AF'		;get error flag
		RET

;END OF TP-RTL3.INC
;------------------
