;--------------------------------
;TP-RTL4.INC	ho	03-MAR-94
;--------------------------------
;V1.00: 0A81h-0F37h
;V2.00: 0ABEh-0F74h	V2.01: 0AB8h-0F6Eh
;V3.00: 0AF5h-0FABh	V3.01: 0AF6h-0FACh

		IDENT	TP-RTL4.INC

;X / Y Real (internal)
;E: X in BCDEHL, Y in BCDEHL'
;V: AF, BC, DE, HL, AF'
;R: Carry if overflow
;   else X / Y in BCDEHL (Y = BCDEHL' unchanged!)
@DivReal:	LD	A,L		;X = 0.0?
		OR	A
		RET	Z		;yes -> do nothing
		EXX
		SUB	L		;subtract charact.
		EXX
		CCF			;correct sign of exponent of result
		CALL	MulDivVor
		PUSH	HL		;reserve space on stack
		PUSH	HL		;for result of division
		PUSH	HL
		ADD	IX,SP		;pointer to result
		EXX
		LD	L,5		;byte counter in char. byte of Y
		EXX			;(charact. is already evaluated)
		LD	A,8		;bit counter
DivReal1:	EX	AF,AF'		;save counter
		CALL	CompMantisse	;does Y fit in?
		JR	C,DivReal2	;no -> shift 0 into result
		CALL	SubMantisse	;yes, subtract and shift 1 into it
DivReal2:	CCF
		RL	L		;shift one bit into result
		EX	AF,AF'		;get bit counter
		DEC	A		;finished?
		JR	NZ,DivReal3	;no -> next bit
		LD	(IX+05),L	;store one byte of result
		DEC	IX		;pointer to next byte
		EXX
		DEC	L		;decrease byte counter. Finished?
		EXX
		JR	Z,DivReal4	;yes -> get result
		LD	A,8		;no, reset bit counter
DivReal3:	CALL	SLAMantisse	;shift mantissa. Did a bit fell out?
		JR	NC,DivReal1	;no -> continue
	;yes, now we can subtract without comparing first (big enough)
		EX	AF,AF'		;save bit counter
		CALL	SubMantisse
		OR	A		;clear Carry (put 1 into result)
		JR	DivReal2	;and continue

	;result is now on stack
	;calculate another result bit (for norming)
DivReal4:	CALL	SLAMantisse	;shift mantissa. Did a bit fell out?
		JR	C,DivReal5	;yes -> save it
		CALL	CompMantisse	;no. Compare: does Y fit in?
		CCF			;save it
DivReal5:	POP	HL		;get result
		POP	DE
		POP	BC
	;norm mantissa
		BIT	7,B		;most significant bit = 1?
		JR	NZ,DivReal6	;yes -> adapt charac.
		CALL	RLMantisse	;shift in additional bit
		JR	MulDiv1		;finished! (ok.)

DivReal6:	INC	L		;overflow?
		JR	NZ,MulDiv1	;no -> finished! (ok.)
		SCF			;yes, error!
		JR	MulDiv2

;Preparation for multiplication/division
;E: X in BCDEHL, Y in BCDEHL'
;   mult.: charakteristik of X + ch.Y in A, Carry flag according
;   div.:  ch.X - ch.Y in A, Carry turned around
;V: AF, B, L, B', IX, stack
;R: Carry and return to the last but one routine if overflow,
;   BCDEHL=0, NCarry and return to the last but one routine if underflow,
;   else:
;   exponent of result in L, mantissae of X and Y completed,
;   IX on stack, Y partially on stack, sign of result on stack,
;   IX =0. Contents of stack: AF/BC/HL/IX
MulDivVor:	JR	C,MuDiVor1	;exp. greater or equal 0?
		;no, exp. of result is smaller than 0
		ADD	A,80h		;convert to charact.
		JR	C,MuDiVor2	;ok, exp. wasn't too small
		JR	MuDiVor3	;underflow!

MuDiVor1:	;yes, exp. of result greater or equal 0
		ADD	A,80h		;convert to charact.
		JR	C,MuDiVor3	;overflow!
MuDiVor2:	;save new charact.
		LD	L,A
		EX	(SP),IX		;save IX, pop return address
		EXX
		PUSH	HL		;save Y partially (except DE)
		PUSH	BC
		LD	A,B
		SET	7,B		;complete mantissa of Y
		EXX
		XOR	B		;xor the signs of X and Y
		AND	80h		;gives sign of result
		PUSH	AF		;save it
		SET	7,B		;complete mantissa of X
		PUSH	IX		;restore return address
		LD	IX,0
		RET

MuDiVor3:	;Carry if overflow, NCarry if underflow
		POP	HL		;throw away return address
		RET	C		;return with Carry if overflow
					;return with Null if underflow
;<fall through>

;E:- / V: AF, BC, DE, HL / R: Zero, NCarry, A=BCDEHL=0
LoadZero:	XOR	A
		LD	L,A
		LD	B,A
		LD	C,A
		LD	D,A
		LD	E,A
		LD	H,A
	;00 00 00 00 00 00 = 0.0000000000E+00 = 0.0
		RET

;shift right logical (0 -> B -> C -> D -> E -> H -> Carry)
;E: mantissa in BCDEH
;V: F, BCDEH
;R: Carry, if bit 0 of H was set, BCDEH shifted right
SRLMantisse:	OR	A		;clear Carry
;<fall through>

;rotate right (Carry -> B -> C -> D -> E -> H -> Carry)
;E: mantisse in BCDEH, N/Carry
;V: F, BCDEH
;R: Carry, if bit 0 of H was set, Carry shifted into BCDEH
RRMantisse:	RR	B
		RR	C
		RR	D
		RR	E
		RR	H
		RET

;shift left arithmetically (Carry <- B <- C <- D <- E <- H <- 0)
;E: mantissa in BCDEH
;V: F, BCDEH
;R: Carry, if bit 7 of B was set, BCDEH shifted left
SLAMantisse:	OR	A		;clear Carry
;<fall through>

;rotate left (Carry <- B <- C <- D <- E <- H <- Carry)
;E: mantissa in BCDEH, N/Carry
;V: F, BCDEH
;R: Carry, if bit 7 of B was set, Carry shifted into BCDEH
RLMantisse:	RL	H
		RL	E
		RL	D
		RL	C
		RL	B
		RET

;E: X-mantissa in BCDEH, Y-mantissa in BCDEH'
;V: AF, BCDEH
;R: X-mant. + Y-mant. in BCDEH, N/Carry
AddMantisse:	LD	A,H		;begin with least
		EXX			;significant byte
		ADD	A,H
		EXX
		LD	H,A
		LD	A,E
		EXX
		ADC	A,E
		EXX
		LD	E,A
		LD	A,D
		EXX
		ADC	A,D
		EXX
		LD	D,A
		LD	A,C
		EXX
		ADC	A,C
		EXX
		LD	C,A
		LD	A,B
		EXX
		ADC	A,B
		EXX
		LD	B,A
		RET

;E: X-mantissa in BCDEH, Y-mantissa in BCDEH'
;V: AF, BCDEH
;R: X-mant. - Y-mant. in BCDEH, N/Carry
SubMantisse:	LD	A,H		;begin with least
		EXX			;significant byte
		SUB	H
		EXX
		LD	H,A
		LD	A,E
		EXX
		SBC	A,E
		EXX
		LD	E,A
		LD	A,D
		EXX
		SBC	A,D
		EXX
		LD	D,A
		LD	A,C
		EXX
		SBC	A,C
		EXX
		LD	C,A
		LD	A,B
		EXX
		SBC	A,B
		EXX
		LD	B,A
		RET

;E: X-mantissa in BCDEH, Y-mantissa in BCDEH'
;V: AF
;R: NZero,  Carry if X-mant. < Y-mant.
;    Zero, NCarry if X-mant. = Y-mant.
;   NZero, NCarry if X-mant. > Y-mant.
CompMantisse:	LD	A,B		;begin with least
		EXX			;significant byte
		CP	B
		EXX			;if not equal then
		RET	NZ		;we are already finished
		LD	A,C
		EXX
		CP	C
		EXX
		RET	NZ
		LD	A,D
		EXX
		CP	D
		EXX
		RET	NZ
		LD	A,E
		EXX
		CP	E
		EXX
		RET	NZ
		LD	A,H
		EXX
		CP	H
		EXX
		RET

;E: X in BCDEHL, Y in BCDEHL'
;V: AF
;R: NZero,  Carry if X < Y
;    Zero, NCarry if X = Y
;   NZero, NCarry if X > Y
CompReals:	EXX
		LD	A,B		;do X and Y have the same sign?
		EXX
		XOR	B
		JP	P,CompReals1	;yes -> need to look closer
		LD	A,B		;no, sign of X tells us
		RLA			;(X is smaller than Y if X negative)
		RET

CompReals1:	BIT	7,B		;X positive?
		JR	Z,CompReals2	;yes -> compare
		CALL	CompReals2	;no, have to flip
		RET	Z		;result of comparison
		CCF
		RET

CompReals2:	;compare REAL number ignoring sign
		LD	A,L
		EXX
		CP	L		;characteristics
		EXX
		RET	NZ		;not equal -> finished
		OR	A		;equal, but zero -> finished
		RET	Z		;equal but not zero ->
		JP	CompMantisse	;compare mantissae


;int(X)		return integer part
;E: X in BCDEHL
;V: AF, BCDEHL / R: int(X) in BCDEHL
;Method:
;Cut away bits after decimal point with a mask
IntReal:	LD	A,L
		SUB	81h		;smaller than 2^0=1?
		JP	C,LoadZero	;yes -> then return 0.0
		INC	A
		CP	40		;greater than 2^40? yes -> then the whole
		RET	NC		;mantissa is in front of the decimal point
		EXX			;(i.e. there is no fractional part)
		PUSH	BC		;A contains the number of bits of the mantissa
		PUSH	DE		;that are in front of the decimal point
		PUSH	HL		;save registers
		EX	AF,AF'		;save A
		CALL	LoadZero	;BCDEHL := 0
		EX	AF,AF'		;put together mask:
IntReal1:	SCF
		CALL	RRMantisse	;shift in ones
		DEC	A
		JR	NZ,IntReal1
		EXX			;now cut away fractional
		LD	A,H		;part with mask
		EXX
		AND	H
		EXX
		LD	H,A
		LD	A,E
		EXX
		AND	E
		EXX
		LD	E,A
		LD	A,D
		EXX
		AND	D
		EXX
		LD	D,A
		LD	A,C
		EXX
		AND	C
		EXX
		LD	C,A
		LD	A,B
		EXX
		AND	B
		EXX
		LD	B,A
IntFrac:	JP	PopXReal


;frac(X)	return fractional part
;E: X in BCDEHL
;V: AF, BCDEHL / R: frac(X) in BCDEHL
;Method:
;frac(X) = X - int(X)
FracReal:	EXX
		PUSH	BC		;save BCDEHL'
		PUSH	DE
		PUSH	HL
		EXX
		CALL	DupReal		;duplicate X
		EXX
		CALL	IntReal
		EXX
		CALL	@SubReal
		JR	IntFrac


;sqrt(X)	square root
;E: X in BCDEHL
;V: AF, BCDEHL, BCDEHL'
;R: sqrt(X) in BCDEHL
;Method: babylonic quare root
;a.k.a. "HERON" method (HERON from Alexandria, +-75 n.C., greek mathematician)
; (1)	X0 choose number randomly
; (2)	Xn+1 := 1/2 * (Xn + X/Xn)
; iterate, until precise enough
;Here:
; X0 = guessed square root of X (take number with half the exponent but
; identical mantissa)
; remaining error = Xn+1 - Xn
; error limit = X0/2^20
; this leads to the following stop rule:
; charact. of remainig error smaller or equal to (charact. of X0) - 20
SqrtReal:	LD	A,L
		OR	A		;X =0.0?
		RET	Z		;yes -> result = 0.0
		BIT	7,B		;X negative?
		LD	A,03h		;yes -> Sqrt argument error
		JP	NZ,RunTimErrX
		CALL	DupReal		;X = BCDEHL'
	;calculate X0, the guessed square root
		LD	A,L		;get characteristic
		ADD	A,80h		;convert to exponent
		SRA	A		;divide exp. by 2 (signed)
		ADD	A,80h		;convert back to charact.
		LD	L,A		;save charact.
	;calculate characteristic of stop value
		SUB	20		;char. of X0 - 20
		PUSH	AF
		EXX
	;X is in BCDEHL, Xn in BCDEHL'
SqrtReal1:	PUSH	BC		;save X
		PUSH	DE
		PUSH	HL
	;calculate Xn+1
		CALL	@DivReal	;X/Xn
		CALL	@AddReal	;+ Xn
		DEC	L		;divide by 2
		PUSH	BC		;save Xn+1
		PUSH	DE
		PUSH	HL
		CALL	@SubReal	;calculate remaining error Xn+1 - Xn
		LD	A,L		;save its char.
		POP	HL		;get Xn+1
		POP	DE
		POP	BC
		EXX			;BCDEHL' = Xn+1
		POP	HL		;get X (and throw away Xn)
		POP	DE
		POP	BC
	;precise enough?
		EX	(SP),HL		;charact. of stop value to H
		CP	H		;char. of remaining error bigger
		EX	(SP),HL		;than char. of stop value?
		JR	NC,SqrtReal1	;yes -> continue, not precise enough
		POP	AF		;throw away char. of stop value
		EXX			;result = Xn+1 to BCDEHL
		RET


;cos(X)		Cosinus
;E: X in BCDEHL
;V: AF, BCDEHL, BCDEHL'
;R: cos(X) in BCDEHL
;Method: cos(X) = sin(Pi/2 - X)
CosReal:	EXX			;X to BCDEHL'
		CALL	LoadPi		;Pi to BCDEHL
		DEC	L		;half it
		CALL	@SubReal	;subtract
;<fall through>


;sin(X)		Sinus
;E: X in BCDEHL
;V: AF, BCDEHL, BCDEHL'
;R: sin(X) in BCDEHL
;Method:
; (1)	Exploit symmetries (see TdM p.179).
;	a) period 2*Pi --> use sin(X) = sin(X mod 2*Pi) to limit to 0-2*Pi
;	b) X > Pi --> use sin(X) = -sin(X-Pi) to limit to 0-Pi
;	c) X > Pi/2 --> use sin(X) = sin(Pi-X) to limit to 0-Pi/2
; (2)	With X near zero sin(X) can be well approximated by X itself.
;	With such small values rounding errors of the normal method
;	would do worse than this approximation.
; (3)	Additional limitation of the definition range to 0-Pi/6
;	with the addition theorem:
;	  sin(3*X) = 3*sin(X) + 4*(sin(X)^3) (see TdM p.181)
;	here a somewhat transformed version is used:
;	  sin(X) = (3/4 * sin(X/3) + sin(X/3)^3)/4
; (4)	Series of powers method (see TdM p.32)
;	sin(X) = X - X^3/3! + X^5/5! - ...
;While summing up the value oscillates around the wanted value of sin(X)
;depending on whether we add or subtract. Because the single summands
;are getting smaller everytime we can say that the remaining error is
;a) one-sided, i.e. if we added something, the sum is surely not
;   smaller than sin(X), and
;b) surely not bigger than the first unused summand.
;Because we sum up until -X^11/11! here, our result is not greater than
;sin(X) and at most X^13/13! too small. Because X is at most Pi/6, the
;maximum error is 3.57E-14, that is approx. 2^-44. This means that the
;error is only visible after the first 44 bits. Because our mantissa
;has only 5 byte = 40 bits this means that our result is as exactly
;as our representation allows. So stopping the series after -X^11/11! is OK.
SinReal:	EXX			;2*Pi to BCDEHL'
		CALL	LoadPi
		INC	L
		EXX
		LD	A,L		;X smaller than
		CP	108		;1/2^20?
		RET	C		;yes -> do nothing
		PUSH	BC		;save sign
		RES	7,B		;make absolute
		CALL	CompReals	;smaller than 2*Pi?
		POP	BC
		JR	C,SinReal1	;yes -> already in limited range
	;calculate X modulo 2*Pi (exploit symmetry)
		CALL	@DivReal
		CALL	FracReal
		CALL	@MultReal
SinReal1:	BIT	7,B		;X negative?
		JR	Z,SinReal2	;no -> do nothing
		CALL	@AddReal	;yes, add 2*Pi
SinReal2:	EXX
		DEC	L		;BCDEHL' = Pi
		EXX
		CALL	CompReals	;X > Pi?
		PUSH	AF		;save result (for dealing with sign later)
		JR	C,SinReal3	;no -> do nothing
		CALL	@SubReal	;yes, subtract Pi
SinReal3:	EXX
		DEC	L		;BCDEHL' = Pi/2
		EXX
		CALL	CompReals	;X > Pi/2?
		JR	C,SinReal4	;no -> do nothing
		EXX			;yes, subtract X of Pi
		INC	L
		CALL	@SubReal
SinReal4:	LD	A,L		;X smaller than
		CP	108		;1/2^20?
		JR	C,SinReal5	;yes -> do nothing
	;now divide by 3
		EXX
		LD	BC,2AAAh
		LD	DE,0AAAAh
		LD	HL,0AA7Fh
	;7F AA AA AA AA 2A = 3.3333333333E-01 = 1/3
		CALL	@MultReal
	;Series of powers method
		PUSH	IX
		LD	IX,SinTab-6
		LD	A,5		;number of summands
		CALL	PotReiheU
		POP	IX
	;addition theorem
		CALL	DupReal		;BCDEHL= sin(X/3)
		CALL	@MultReal
		CALL	@MultReal
		PUSH	BC
		PUSH	DE
		PUSH	HL		;save sin(x/3)^3
		EXX
		CALL	DupReal
		DEC	L
		DEC	L		;sin(x/3)/4
		EXX
		DEC	L		;sin(X/3)/2
		CALL	@AddReal	;gives 3/4 * sin(X/3)
		EXX
		POP	HL		;get sin(X/3)^3
		POP	DE
		POP	BC
		EXX
		CALL	@SubReal	;3/4 * sin(X/3) - sin(X/3)^3
		INC	L		;/4 gives sin(X)
		INC	L
SinReal5:	POP	AF		;get whether Pi was greater
		INC	L		;sin(X)=0?
		DEC	L
		RET	Z		;yes -> 0 has no sign
		RET	C		;flip? no -> finished
	;flip sign (wouldn't SET 7,B be sufficient here?)
		LD	A,B
		XOR	80h
		LD	B,A
		RET

;coefficient table for Sinus
SinTab:	DEFB	067h,0AAh,03Fh,02Bh,032h,0D7h	;= -2.5052108385E-08 = -1/11!
	DEFB	06Eh,0B6h,02Ah,01Dh,0EFh,038h	;=  2.7557319224E-06 =  1/9!
	DEFB	074h,00Dh,0D0h,000h,00Dh,0D0h	;= -1.9841269841E-04 = -1/7!
	DEFB	07Ah,088h,088h,088h,088h,008h	;=  8.3333333333E-03 =  1/5!
	DEFB	07Eh,0ABh,0AAh,0AAh,0AAh,0AAh	;= -1.6666666667E-01 = -1/3!


;ln(X)		Logarithmus naturalis
;E: X in BCDEHL
;V: AF, BCDEHL, BCDEHL'
;R: ln(X) in BCDEHL
;Method:
; (1)	Move into range 0..1:
;	X= (2^A) * B, where A integer number and B between 0 and 1
;	ln(X) = ln((2^A)*B) = ln(2^A) + ln(B) = A * ln(2) + ln(B)
; (2)	ln(X) = ln(X/sqrt(2)) + 1/2 * ln(2)
; (3)	Series of powers method (see TdM p.33)
;	ln(X) = 2 * ((X-1)/(X+1) + (X-1)^3/3*(X+1)^3 + (X-1)^5/5*(X+1)^5 + ...)
LnReal:		INC	L		;charact of X=0,
		DEC	L		;i.e. X=0?
		LD	A,04h		;yes -> LN argument error
		JP	Z,RunTimErrX
		BIT	7,B		;negative?
		JP	NZ,RunTimErrX	;yes -> error too
		EXX
		CALL	LoadSqu2
		EXX
		LD	A,L
		LD	L,81h		;move into range 0..1
		SUB	L		;calculate A
		PUSH	AF		;and save it
		CALL	@DivReal	;divide B by square root of 2
		EXX
		CALL	LoadOne
		EXX
		CALL	@SubReal
		PUSH	BC		;save B/sqrt(2)-1
		PUSH	DE
		PUSH	HL
		EXX
		INC	L		;times 2 = 2.0
		CALL	@AddReal	;B/sqrt(2)+1
		EXX
		POP	HL
		POP	DE
		POP	BC
		CALL	@DivReal	;=(B/sqrt(2)-1)/(B/sqrt(2)+1)
	;Serie of powers method
		PUSH	IX
		LD	IX,LnTab-6
		LD	A,6		;number of summands
		CALL	PotReiheU
		POP	IX
		INC	L		;times 2 = ln(B/sqrt(2))
		EXX
		CALL	LoadLn2
		DEC	L		;1/2 * ln(2)
		EXX
		CALL	@AddReal	;+ = ln(B)
		POP	AF		;get A
		PUSH	BC
		PUSH	DE
		PUSH	HL
		LD	L,A		;extend A signed
		LD	H,0		;to an integer
		JR	NC,LnReal1
		DEC	H
LnReal1:	CALL	Int2Real	;extend to real
		EXX
		INC	L		;times 2 = ln(2)
		CALL	@MultReal	;= A * ln(2)
		EXX
		POP	HL		;get ln(B)
		POP	DE
		POP	BC
		CALL	@AddReal	;= ln(X)
		LD	A,L
		CP	67h		;smaller than 2^-25?
		JP	C,LoadZero	;yes -> return zero instead
		RET

;coefficient table for Ln
LnTab:	DEFB	07Dh,08Ah,09Dh,0D8h,089h,01Dh	;= 7.6923076923E-02 = 1/13
	DEFB	07Dh,0E9h,0A2h,08Bh,02Eh,03Ah	;= 9.0909090909E-02 = 1/11
	DEFB	07Dh,08Eh,0E3h,038h,08Eh,063h	;= 1.1111111111E-01 = 1/9
	DEFB	07Eh,049h,092h,024h,049h,012h	;= 1.4285714286E-01 = 1/7
	DEFB	07Eh,0CDh,0CCh,0CCh,0CCh,04Ch	;= 2.0000000000E-01 = 1/5
	DEFB	07Fh,0ABh,0AAh,0AAh,0AAh,02Ah	;= 3.3333333333E-01 = 1/3


;exp(X)		Exponential function
;E: X in BCDEHL
;V: AF, BCDEHL, BCDEHL'
;R: exp(X) in BCDEHL
;Method:
; (1)	if X negative, then ^X = 1/e^abs(x)
; (2)	?? if anybody knows exactly whats going on, let me know!!
ExpReal:	EXX
		CALL	LoadLn2
		EXX
		OR	A
		BIT	7,B		;X negative?
		PUSH	AF		;save it
		RES	7,B		;take absolute value
		CALL	@DivReal
		LD	A,L
		CP	88h
		JR	NC,ExpReal3	;Overflow
		PUSH	BC
		PUSH	DE
		PUSH	HL
		INC	L
		CALL	RoundReal
		PUSH	HL
		SRL	H
		RR	L
		LD	A,L
		POP	HL
		PUSH	AF
		CALL	Int2Real
		INC	L
		DEC	L
		JR	Z,ExpReal1
		DEC	L
ExpReal1:	EXX
		POP	AF
		POP	HL
		POP	DE
		POP	BC
		PUSH	AF
		CALL	@SubReal
		PUSH	IX
		LD	IX,ExpTab-6
		LD	A,8		;number of summands
		CALL	PotReihe
		POP	IX
		POP	AF
		JR	NC,ExpReal2
		PUSH	AF
		EXX
		CALL	LoadSqu2
		EXX
		CALL	@MultReal
		POP	AF
ExpReal2:	ADD	A,L
		LD	L,A
		JR	C,ExpReal3	;Overflow
		POP	AF		;get whether X was negative
		RET	Z		;no -> finished
		EXX			;yes, invert
		CALL	LoadOne
		JP	@DivReal

ExpReal3:	POP	HL		;throw away saved AF
		LD	A,01h		;Floating point overflow
		JP	RunTimErrX

;coefficient table for Exp
ExpTab:	DEFB	06Dh,02Eh,01Dh,011h,060h,031h	;= 1.3215486790E-06 = ln(2.0)^8/8!
	DEFB	070h,046h,02Ch,0FEh,0E5h,07Fh	;= 1.5252733804E-05 = ln(2.0)^7/7!
	DEFB	074h,036h,07Ch,089h,084h,021h	;= 1.5403530393E-04 = ln(2.0)^6/6!
	DEFB	077h,053h,03Ch,0FFh,0C3h,02Eh	;= 1.3333558146E-03 = ln(2.0)^5/5!
	DEFB	07Ah,0D2h,07Dh,05Bh,095h,01Dh	;= 9.6181291076E-03 = ln(2.0)^4/4!
	DEFB	07Ch,025h,0B8h,046h,058h,063h	;= 5.5504108665E-02 = ln(2.0)^3/3!
	DEFB	07Eh,016h,0FCh,0EFh,0FDh,075h	;= 2.4022650696E-01 = ln(2.0)^2/2!
	DEFB	080h,0D2h,0F7h,017h,072h,031h	;= 6.9314718056E-01 = ln(2.0)


;arctan(X)	Arcustangens
;E: X in BCDEHL
;V: AF, BCDEHL, AF', BCDEHL'
;R: arctan(X) in BCDEHL
;Method:
; (1)	arctan(X) = -arctan(-X), limitation to positive values
; (2)	if X > tan(Pi/24)
;	Shift the range (segmented, in three segments), using formula
;	arctan(X)-arctan(Y) = arctan((X-Y)/(1+X*Y)) (see TdM p.185)
;	with Y taken from a table. Premise: X*Y > -1
; (3)	if X < 1.0, then arctan(X) = X - X^3/3 + X^5/5 - X^7/7 ...
;	else arctan(X) = Pi/2 - 1/X + 1/3*X^3 - 1/5*X^5 + ... (see TdM p.34)
AtanReal:	LD	A,L		;X=0?
		OR	A
		RET	Z		;yes -> result is 0
		PUSH	IX
		EXX
		CALL	LoadOne		;1.0
		EXX
		XOR	A
		BIT	7,B		;X negative?
		JR	Z,AtanReal1
		INC	A		;yes ->set flag 'negative'
		RES	7,B		;and take absolute value
AtanReal1:	PUSH	AF		;save sign flag
		CALL	CompReals	;X < 1.0?
		JR	C,AtanReal2	;yes -> we need X^...
		EXX			;no, we need 1/X^...
		CALL	@DivReal	;invert it
		POP	AF
		SET	7,A		;set flag for 'absolute value >= 1.0'
		PUSH	AF

AtanReal2:	;if X < tan(Pi/24), then don't shift the range
		EXX
		LD	BC,06CFh
		LD	DE,0E98Eh
		LD	HL,4A7Eh
	;7E 4A 8E E9 CF 6C = 1.3165249759E-01 = tan(Pi/24)
		EXX
		CALL	CompReals	;smaller?
		JR	NC,AtanReal3	;no -> shift range
		CALL	AtanReihe	;yes, calculate directly
		JR	AtanReal6

AtanReal3:	;shift range
		LD	IX,AtanT1-18
		LD	A,2
AtanReal4:	EX	AF,AF'		;save counter
		EXX
		LD	DE,18		;advance pointer 3 numbers
		ADD	IX,DE
		CALL	GetReal		;read number
		EXX
		CALL	CompReals	;X < this number?
		JR	C,AtanReal5	;yes -> 
		EX	AF,AF'		;no, continue searching
		DEC	A
		JR	NZ,AtanReal4
		EXX
		LD	DE,12		;neither, so take last
		ADD	IX,DE		;entry of table
		EXX
AtanReal5:	EXX
		CALL	GetNxtReal	;get Y
		SET	7,B		;make negative
		CALL	@AddReal	;add
		PUSH	BC		;save X-Y
		PUSH	DE
		PUSH	HL
		CALL	GetReal		;get Y again
		CALL	@MultReal	;BCDEHL = X * Y
		EXX
		CALL	LoadOne
		CALL	@AddReal	;BCDEHL = 1+X*Y
		EXX
		POP	HL		;X-Y holen
		POP	DE
		POP	BC
		CALL	@DivReal	;BCDEHL = (X-Y)/(1+X*Y)
		PUSH	IX		;save pointer in table
		CALL	AtanReihe	;BCDEHL = arctan((X-Y)/(1+X*Y))
		POP	IX		;get pointer
		EXX
		CALL	GetNxtReal	;get arctan(Y)
		CALL	@AddReal	;and add

AtanReal6:	;now arctan(X) is almost calculated
		POP	AF		;get flags
		RLA			;was absolute value >= 1.0?
		JR	NC,AtanReal7	;no -> result of series ok
		PUSH	AF		;yes, calculate Pi/2 minus
		EXX			;result of the series
		CALL	LoadPi
		DEC	L
		CALL	@SubReal
		POP	AF

AtanReal7:	;arctan(X) is now correct except sign
		POP	IX		;restore IX
		BIT	1,A		;was X negative?
		RET	Z		;no -> finished
		SET	7,B		;yes, then result negative
		RET

;Values for range shift for Arcustangens
AtanT1:	;Range I:	0 < X < tan(Pi/24)
	;	is not shifted

	;Range II:	X < sqrt(2)-1
	;	arctan(X) = arctan((X-tan(Pi/12))/(1+X*tan(Pi/12))) + Pi/12
	DEFB	07Fh,0E7h,0CFh,0CCh,013h,054h	;=  4.1421356237E-01 = sqrt(2)-1
	DEFB	07Fh,0F6h,0F4h,0A2h,030h,009h	;=  2.6794919243E-01 = tan(Pi/12)
	DEFB	07Fh,06Ah,0C1h,091h,00Ah,006h	;=  2.6179938780E-01 = Pi/12

	;Range III:	X < ???
	;	arctan(X) = arctan((X-tan(Pi/6))/(1+X*tan(Pi/6))) + Pi/6
	DEFB	080h,0B5h,09Eh,08Ah,06Fh,044h	;=  7.6732698798E-01 = ???
	DEFB	080h,082h,02Ch,03Ah,0CDh,013h	;=  5.7735026919E-01 = tan(Pi/6)
	DEFB	080h,06Ah,0C1h,091h,00Ah,006h	;=  5.2359877560E-01 = Pi/6

	;Range IV:	X >= ???
	;	arctan(X) = arctan((X-tan(Pi/4))/(1+X*tan(Pi/4))) + Pi/4
	DEFB	081h,000h,000h,000h,000h,000h	;=  1.0000000000E+00 = tan(Pi/4)
	DEFB	080h,021h,0A2h,0DAh,00Fh,049h	;=  7.8539816340E-01 = Pi/4

;coefficient table for Arcustangens
AtanTab:DEFB	07Dh,0E8h,0A2h,08Bh,02Eh,0BAh	;= -9.0909090909E-01 = -1/11
	DEFB	07Dh,08Eh,0E3h,038h,08Eh,063h	;=  1.1111111111E-02 =  1/9
	DEFB	07Eh,049h,092h,024h,049h,092h	;= -1.4285714286E-01 = -1/7
	DEFB	07Eh,0CDh,0CCh,0CCh,0CCh,04Ch	;=  2.0000000000E-01 =  1/5
	DEFB	07Fh,0ABh,0AAh,0AAh,0AAh,0AAh	;= -3.3333333333E-01 = -1/3

AtanReihe:	LD	IX,AtanTab-6
		LD	A,5
;<fall through>

;Series of powers method (use only odd powers of X)
;E: X in BCDEHL, A number n of table entries,
;   IX points in front of coefficient table (Kn, Kn-1, Kn-2, ...)
;V: AF, BCDEHL, AF', BCDEHL'
;R: BCDEHL = X + K1 * X^3 + K2 * X^5 + K3 * X^7 ...
PotReiheU:	PUSH	BC		;save X
		PUSH	DE
		PUSH	HL
		PUSH	AF		;save counter
		CALL	DupReal
		CALL	@MultReal	;square X
		POP	AF		;get counter
		CALL	PotReihe
		EXX			;free up BCDEHL
		POP	HL		;for saved X
		POP	DE
		POP	BC
		JP	@MultReal

;Series of powers method (all powers of X)
;E: X in BCDEHL, A number n of table entries,
;   IX points in front of coefficient table (Kn, Kn-1, Kn-2, ...)
;V: AF, BCDEHL, AF', BCDEHL'
;R: BCDEHL = (...((Kn * X + Kn-1) * X + Kn-2) *  ...+ K1) * X + 1
;          = 1 + K1 * X + K2 * X^2 + K3 * X^3 + ...
PotReihe:	PUSH	AF		;save sounter
		EXX			;save X 
		CALL	GetNxtReal	;get Kn
		JR	PotReihe2

PotReihe1:	PUSH	AF		;save counter again
		EXX			;away with temporary result, get X
		PUSH	BC		;save X
		PUSH	DE
		PUSH	HL
		CALL	GetNxtReal	;get next constant
		CALL	@AddReal	;BCDEHL = new temporary result
		EXX			;stow away
		POP	HL		;old temporary result ist
		POP	DE		;replaced by saved X
		POP	BC
		EXX			;swap, so X is not destroyed
PotReihe2:	CALL	@MultReal	;temporary result * X
		POP	AF		;get counter
		DEC	A		;finished?
		JR	NZ,PotReihe1	;no -> add next constant
		EXX			;yes, add 1
		CALL	LoadOne
		JP	@AddReal

;E: pointer to real number last used in IX
;V: F, BCDEHL, IX
;R: next real number in BCDEHL, pointer to it in IX
GetNxtReal:	LD	DE,6		;SizeOf(REAL)
		ADD	IX,DE
;<fall through>

;E: pointer to real number in IX
;V: BCDEHL
;R: real number in BCDEHL
GetReal:	LD	L,(IX)		;LowByte first
		LD	H,(IX+01)
		LD	E,(IX+02)
		LD	D,(IX+03)
		LD	C,(IX+04)
		LD	B,(IX+05)
		RET

;return constant 1.0
;E:- / V: BCDEHL / R: BCDEHL = 1.0
LoadOne:	LD	HL,81h
		LD	B,H
		LD	C,H
		LD	D,H
		LD	E,H
	;81 00 00 00 00 00 = 1.0000000000E+00 = 1.0
		RET

;return constant Pi
;E:- / V: BCDEHL / R: BCDEHL = Pi
LoadPi:		LD	BC,490Fh
		LD	DE,0DAA2h
		LD	HL,2182h
	;82 21 A2 DA 0F 49 = 3.1415926536E+00 = Pi
		RET

;return constant square root of 2
;E:- / V: BCDEHL / R: BCDEHL = square root of 2
LoadSqu2:	LD	BC,3504h
		LD	DE,0F333h
		LD	HL,0FA81h
	;81 FA 33 F3 04 35 = 1.4142135624E+00 = sqrt(2.0)
		RET

;return constant Ln 2
;E:- / V: BCDEHL / R: BCDEHL =  Ln 2
LoadLn2:	LD	BC,3172h
		LD	DE,17F7h
		LD	HL,0D280h
	;80 D2 F7 17 72 31 = 6.9314718056E-01 = ln(2.0)
		RET

;END OF TP-RTL4.INC
;------------------
