;--------------------------------
;TP-RTL5.INC	ho	11-APR-94
;--------------------------------
;V1.00: 0F38h-12FAh
;V2.00: 0F75h-133Bh	V2.01: 0F6Fh-1335h
;V3.00: 0FACh-136Eh	V3.01: 0FADh-136Fh

		IDENT	TP-RTL5.INC

;duplicate X
;E: X in BCDEHL
;V: BCDEHL'
;R: X also in BCDEHL'
DupReal:	PUSH	BC
		PUSH	DE
		PUSH	HL
		EXX
		POP	HL
		POP	DE
		POP	BC
		RET

;random()	random number
;E:- / V: AF, BCDEHL
;R: random number (0 <= random < 1) in BCDEHL
RandReal:	CALL	Random		;32 bit random number in BCDE
		;expand mantissa with 00h (reg. H),
		;characteristic = 80h (reg. L), i.e. range 0-1
		LD	HL,80h
		LD	A,32
		;norm it: shift left first 32 bit of
		;mantissa until first bit is a 1.
RandReal1:	BIT	7,B
		JR	NZ,RandReal2
		SLA	E
		RL	D
		RL	C
		RL	B
		DEC	L
		DEC	A
		JR	NZ,RandReal1
		;if mantissa was all zeros, return 0.0
		LD	L,A		;characteristic := 0
RandReal2:	RES	7,B		;sign: positive
		RET

;round(X)	round
;E: X in BCDEHL
;V: AF, BCDEHL, BCDEHL'
;R: round(X) in HL
;Method:
;round(X) = trunc(X + 1/2)	if X positive
;round(X) = trunc(X - 1/2)	if X negative
RoundReal:	BIT	7,B		;negative?
		EXX
		CALL	LoadOne
		JR	Z,RoundReal1
		SET	7,B		;yes -> minus 1
RoundReal1:	DEC	L		;+- 1/2
		CALL	@AddReal
;<fall through>

;trunc(X)	truncate
;E: X in BCDEHL
;V: AF, BCDEHL
;R: trunc(X) in HL
;Method:
;Shift right mantissa until first 15 bits of mantissa are the integer
;part of the number, insert sign and return in HL.
TruncReal:	OR	A
		;charact. of X less than 128, i.e. exp. of X negative,
		BIT	7,L		;i.e. absolute value of X less than 1?
		JR	Z,TruncReal3	;yes -> return 0
		BIT	7,B		;X negative?
		EX	AF,AF'		;save
		SET	7,B		;complete mantissa
		;shift right mantissa incrementing characteristic,
		;until upper 15 bit of mantissa are integer part of X
TruncReal1:	LD	A,8Fh
		CP	L		;X greater than 2^15?
		JR	C,TruncReal4	;yes -> Out of integer range
		JR	Z,TruncReal2	;finished? -> convert to int
		CALL	SRLMantisse
		INC	L
		JR	TruncReal1

TruncReal2:	CALL	SRLMantisse	;shift once more for sign
		EX	AF,AF'		;get sign
		LD	H,B		;move upper 16 bit of
		LD	L,C		;mantissa to HL
		RET	Z		;positive? yes -> finished
		JP	NegateHL	;no -> flip sign

TruncReal3:	LD	HL,0		;return integer zero
		RET

TruncReal4:	LD	A,92h		;Out of integer range
		JP	RunTimErrX

;convert integer to real
;E: integer in HL
;V: AF, BCDEHL
;R: real number in BCDEHL
;ATTENTION:
;in TPVers1 and TPVers3 there is a severe bug!!
;a call to this routine with HL=$8000 (-32768) hangs in an endless loop!!!
Int2Real: ;special case 0
		LD	A,H		;=0?
		OR	L
		JP	Z,LoadZero	;yes -> return 0.0
		BIT	7,H		;negative?
		EX	AF,AF'		;save it
		CALL	AbsoluteHL	;operate with absolute value first
	;remember: absolute value of $8000 is $8000!
	;this is the only case where we have the most significant bit set
		LD	A,90h		;charact. if HL would start with a 1
IF TPVers2
	;special case $8000
		BIT	7,H
		JR	NZ,Int2Real2
	;oh boy, why the hell did somebody "optimize" that in TPVers3???
ENDIF
	;norm it: shift left, until a 1 reaches the leftmost bit
	;decrementing the characteristic. We don't need to check the first
	;time because the current high-order bit is just the sign which we
	;already treated.
Int2Real1:	ADD	HL,HL		;shift left HL
		DEC	A		;decrement characteristic
		BIT	7,H		;most significant bit =1?
		JR	Z,Int2Real1	;no -> continue
	;ok, the first half of the mantissa and the characteristic
	;are almost finished.
Int2Real2:	LD	B,H		;move it to BC
		LD	C,L
		LD	DE,0		;expand mantissa
		LD	H,D
		LD	L,A		;insert characteristic
	;now process sign
		EX	AF,AF'		;negative?
		RET	NZ		;yes -> finished (sign bit is already 1)
		RES	7,B		;no -> sign := 0
		RET


;convert real -> ascii representation
;E: field width in DE, desired number of digits behind decimal point in
;   HL (-1 if exponential form wanted),
;   pointer to buffer in IX, REAL number in BCDEHL'
;V: AF, BC, DE, HL, IX, IY, BCDEHL'
;R: number in ASCII representation in buffer, pointer behind last character in IX
;Method:
; (1)	convert binary exponent to decimal exponent and multiply the REAL number
;	with 10^-decimalexponent (i.e. norm mantissa to range 1.0 - 9.9...)
; (2)	convert mantissa to decimal ascii representation with maximum precision
;	(12 digits)
; (3)	round mantissa according to desired number of digits
; (4)	process sign, decimal point, exponent, leading zeroes, ...
;	depending on desired representation and write result into buffer
;(1) and (2) are done in subroutine ConvDez, (3) in subroutine RoundDez.
;Only (4) is done here in RtoA.
RtoA:		CALL	Range256	;HL -> A (precision)
		EX	DE,HL		;A=0 and Carry if HL was negative
		LD	E,0		;flag := 0
		JR	C,RtoA1		;chose exponential repres. if precision
		CP	25		;negative or precision greater than 24
		JR	C,RtoA5
RtoA1:	;exponential representation: set flag and calculate precision
	;from field width
		DEC	E		;flag := 0FFh
		CALL	Range256	;HL -> A (field width)
		EXX
		BIT	7,B		;is the REAL number negative?
		EXX
		LD	D,7		;x.mmmmmmmE+xx -> total length =
		JR	Z,RtoA2		;precision plus 6
		INC	D		;negative -> reserve another one for sign
RtoA2:		SUB	D		;= precision
		JR	NC,RtoA3
		XOR	A		;must not be zero
RtoA3:		CP	9
		JR	C,RtoA4
		LD	A,9		;and not more than ten
RtoA4:		INC	A
RtoA5:	;now we have precision in A
		LD	D,A
		PUSH	DE		;save precision and flag
		EXX			;get REAL number
		LD	IY,ConvBuf
		PUSH	IX		;save pointer to buffer
		CALL	ConvDez		;convert raw
		POP	IX
		POP	DE
		LD	C,A		;save decimal exponent
		LD	A,D
		INC	A
		BIT	7,E		;exponential representation desired?
		JR	NZ,RtoA7	;yes ->
	;Look if number is so small that there are more zeros after the
	;decimal point than our precision is. If so, just output a zero.
		ADD	A,C
		JP	P,RtoA6
		LD	(IY),0		;terminate number after the first 0
		JR	RtoA8

RtoA6:	;we now have the number of digits of the calculated digit-string
	;in A that will end up in the output
		CP	12		;less than 12?
		JR	C,RtoA7		;yes -> ok
		LD	A,11		;no, maximum 11
RtoA7:		PUSH	DE
		CALL	RoundDez	;round
		POP	DE
RtoA8:		BIT	7,B		;number positive?
		JR	Z,RtoA9		;yes -> no sign
		LD	A,'-'
		CALL	PutIX
RtoA9:		BIT	7,E		;exponential representation?
		JR	Z,RtoA10	;no ->
		LD	H,C		;decimal exponent
		LD	C,0		;flag := 0
RtoA10:		BIT	7,C
		JR	Z,RtoA11
		CALL	PutIXZero
		JR	RtoA12

RtoA11:	;copy C digits before decimal point
		CALL	PutIXIY
		DEC	C
		JP	P,RtoA11
RtoA12:		LD	A,D		;precision
		OR	A		;none?
		JR	Z,RtoA15	;then no point needed
		LD	A,'.'
		CALL	PutIX
RtoA13:		INC	C		;insert -C-1 zeroes behind point
		JR	Z,RtoA14
		CALL	PutIXZero
		DEC	D
		JR	NZ,RtoA13
RtoA14:	;insert D digits behind point
		DEC	D
		JP	M,RtoA15
		CALL	PutIXIY
		JR	RtoA14

RtoA15:	;mantissa finished
		BIT	7,E		;exponential representation?
		RET	Z		;no -> finished
		LD	A,'E'
		CALL	PutIX
		LD	A,'+'
		BIT	7,H		;exponent positive?
		JR	Z,RtoA16	;yes -> insert '+'
		LD	A,H		;no -> negate exponent
		NEG
		LD	H,A
		LD	A,'-'		;and insert '-'
RtoA16:		CALL	PutIX
		LD	A,H
		LD	B,'0'-1
RtoA17:		INC	B
		SUB	10
		JR	NC,RtoA17	;exp. DIV 10 to B
		ADD	A,'0'+10	;exp. MOD 10 to A
		LD	(IX),B		;insert exponent
		INC	IX
		JR	PutIX

;copy digit from IY to IX
;if (IY)=0, then insert '0'
;else insert (IY) and increment IY
;E: IY points in ConvBuf, IX in buffer
;V: AF, IX, IY / R:-
PutIXIY:	LD	A,(IY)
		INC	IY
		OR	A
		JR	NZ,PutIX
		DEC	IY
;<fall through>

;insert zero
;E:- / V: AF, IX / R: A= 30h, IX incremented
PutIXZero:	LD	A,'0'
;<fall through>

;insert character
;E: character in A, pointer in IX
;V: IX / R: IX incremented
PutIX:		LD	(IX),A
		INC	IX
		RET

;convert number to decimal ascii representation
;(raw, i.e. without decimal point, without exponent, 12 digits)
;E: pointer to ConvBuf in IY, REAL number in BCDEHL
;V: AF, AF', B, DEHL, BCDEHL', IX
;R: decimal exponent in A, sign in bit 7 of B remains
ConvDez:	PUSH	IY		;save
		INC	L
		DEC	L		;number =0.0?
		JR	NZ,ConvDez2	;no -> continue
	;special case 0.0: put 12 zeros into ConvBuf
		LD	B,12
ConvDez1:	LD	(IY),'0'
		INC	IY
		DJNZ	ConvDez1	;B =0 -> zero is positive
		XOR	A		;dec. exponent = 0
		JP	ConvDez8	;leave

ConvDez2: ;normal case
.COMMENT %----------------------------------------------------------------
Method for converting exponent from binary to decimal:

decimal exp. := binary exp. * c, where c = ln 2 / ln 10 = 0.30102...

To speed up, this implementation uses integer math c = 77 / 256 = 0.30078...
This is a bit lower. This means: positive exponents will get a bit too
small, negative exponents a bit too big. Such a deviation on two sides
is not neat. By adding a suitable constant, we can shift the range so
that the deviation occurs only on one side, which simplifies a lot.

Now what constant do we need? The bigger the exponent, the bigger the
deviation. The highest expected deviation for positive binary exponents
is for 127, where we have 127 * ((ln 2 / ln 10) - 77/256) = 0.0315.. which
is about 8/256 (Note that this is well below 1, so that the maximum error
we have to expect for the decimal exponent is 1 less). But here, instead
of 8/256 only 5/256 are added. Why?

Well, 0.0315... is the result of a linear equation, but what we have here
in fact is a diophantic equation, because we don't calculate with real
number precision but with integer math. And we want to keep the constant
as small as possible to increase the deviation on the side of the negative
exponents by an amount as small as possible. As usual, diophantic equations
can only be solved by trying out possible values. We have to find the biggest
binary exponent where multiplication with 77 and division by 256 (which is
simply done by truncation here) would lead to a wrong (i.e. by 1 smaller)
decimal exponent, and then determine the number we have to add before the
truncation to correct that. This number is the constant we need to add to
have deviations on only one side of the range.

Now we elimited the deviation for positive exponents, but what does this
change for negative exponents? The maximum deviation is expected for -128,
where we have 128 * ((ln 2 / ln 10) - 77/256) + 5/256 = 0.0513.. which is
also well below 1, so that the maximum error we have to expect is still
only 1 too much.

Now the situation after calculating ((BinaryExponent * 77) + 5) DIV 256
is that this value is never smaller than the decimal exponent we are
searching, but sometimes (rarely) it is exactly 1 too much.

Did you understand everything? NO?? Don't panic - it's a bit tricky
and I needed quite some time to figure it out for myself ;-)
----------------------------------------------------------------% ;COMMENT
		PUSH	BC		;save sign
		RES	7,B		;make absolute
		LD	A,L		;get characteristic
		EXX			;save absolute value
		SUB	80h		;convert to exponent
		LD	L,A
		SBC	A,A		;extend signed to word
		LD	H,A
		LD	DE,77		;calculate to decimal exponent (raw)
		CALL	MultInt
		LD	DE,5		;shift range to make deviation one-sided
		ADD	HL,DE
		LD	A,H		;now decimal exp. in A, maybe 1 too much
		CP	0D9h		;^-39?
		JR	NZ,ConvDez3	;no -> continue
		INC	A		;yes, one more
	;This is especially for very small numbers. I don't quite
	;get this, but it is probably something like:
	; m.m x 10^-39 --> 0.mm x 10^-38
ConvDez3:	LD	(IY),A		;insert decimal exponent
	;Multiply the REAL number with 10^-DecimalExponent, i.e.
	;norm the number to the range 1.0 - 9.99... If we had a
	;decimal exponent that was one too much, we notice it now
	;that we get a number in the range 0.1 - 0.99... instead.
	;But to correct that now is no big deal.
		NEG			;multiply REAL number (in BCDEHL')
		CALL	Exp10		;with 10^-dec.exp.
		LD	A,L
		CP	81h		;is the result >= 1?
		JR	NC,ConvDez4	;yes -> we made no error
		CALL	Mult10		;no, we have to correct that now
		DEC	(IY)
ConvDez4: ;The decimal digit in front of the decimal point is now
	;in the range 1-9. For the mantissa, which is still a REAL number
	;in registers BCDEHL, this means that its characteristic is between
	;81h (1) and 84h (8-9). To facilitate conversion to decimal, we
	;now norm the mantissa again to a characteristic of 84h.
		SET	7,B		;complete mantissa
		LD	A,84h		;that's what we want
		SUB	L		;is it already?
		LD	L,0		;use L too for the mantissa, so that we
					;don't loose bits when shifting right now
		JR	Z,ConvDez6	;yes -> continue
ConvDez5:	CALL	SRLMantisse	;no, shift right
		RR	L
		DEC	A
		JR	NZ,ConvDez5
ConvDez6: ;Now we have the decimal mantissa as a normed binary number
	;with max. 48+3 bits in BCDEHL. The first 4 bits of B contain
	;the first digit!
		LD	A,(IY)		;get decimal exponent
		PUSH	AF		;and save it
		LD	A,12		;convert dec.mantissa to 12 decimal digits
ConvDez7:	EX	AF,AF'		;save counter
		LD	A,B		;get most significant digit
		RRA
		RRA
		RRA
		RRA
		AND	0Fh
		ADD	A,'0'		;convert to ASCII
		LD	(IY),A		;and insert it
		INC	IY
		LD	A,B		;remove digit
		AND	0Fh
		LD	B,A
		PUSH	BC		;save binary number
		PUSH	DE
		PUSH	HL
		SLA	L		;shift left 2 times
		CALL	RLMantisse	;= multiplication by 4
		SLA	L
		CALL	RLMantisse
		EX	DE,HL		;now add saved number
		EX	(SP),HL		;= mult. by 5
		ADD	HL,DE
		POP	DE
		EX	(SP),HL
		ADC	HL,DE
		EX	DE,HL
		POP	HL
		EX	(SP),HL
		ADC	HL,BC
		LD	B,H
		LD	C,L
		POP	HL
		SLA	L		;shift left another time
		CALL	RLMantisse	;= mult. by 10 --> next digit ready
		EX	AF,AF'		;get counter
		DEC	A
		JR	NZ,ConvDez7	;do next digit
		POP	AF		;saved decimal exponent
		POP	BC		;saved sign in bit 7 of B
ConvDez8:	POP	IY		;pointer to ConvBuf
		RET

;round
;E: pointer to ConvBuf (filled with digit-string) in IY,
;   round after A digits, decimal exponent in C
;V: AF, C, DE, HL
;R: C incremented if overflow when rounding
RoundDez:	PUSH	IY
		POP	HL
		LD	E,A		;address first digit
		LD	D,0		;after the last valid
		ADD	HL,DE
		LD	A,(HL)		;get next and terminate
		LD	(HL),0		;number there
		CP	'5'		;smaller than five?
		RET	C		;yes -> round down, finished
RoundDez1: ;increment previous digit
		DEC	E		;still a digit left?
		JP	M,RoundDez2	;no -> overflow
		DEC	HL
		LD	A,(HL)		;increment digit
		INC	A
		LD	(HL),A
		CP	'9'+1		;was it smaller than nine?
		RET	C		;yes -> finished
		LD	(HL),0		;no, abort number (cut away trailing 0s)
		JR	RoundDez1

RoundDez2: ;overflow: no more digits to increment
		LD	(HL),'1'	;return a 1.0
		INC	HL
		LD	(HL),0
		INC	C		;and increment decimal exponent
		RET


;convert ascii representation -> real number
;E: pointer to string in IX
;V: AF, BCDEHL, BCDEHL', IX
;R: Carry if error and IX points to unrecognized character
;   else number in BCDEHL
;Attention: converts only positive numbers. Unary minus has to be taken
;care of elsewhere!
;Method:
; (1)	read in whole mantissa as a integer number (!),
;	only save position X of the decimal point
; (2)	multiply by 10^(-X) to get the actual mantissa
; (3)	if exponent E was given, convert it to binary and
;	multiply mantissa with 10^E
;FlagByte:
;Bit	76543210
;	   +-------	0= in mantisse, 1= in exponent
;	  +--------	0= positive exponent, 1= negative exp.
;	 +---------	0= in front of decimal point, 1= after dec.point
AtoR:		EXX
		LD	BC,0		;B'= flags, C'=position of decimal point
		EXX
		CALL	LoadZero	;start with 0.0
AtoR1:		LD	A,(IX)		;get char
		CALL	ToUpper
		CP	'.'		;decimal point?
		JR	NZ,AtoR3	;no -> continue
	;decimal point
		EXX
		BIT	6,B		;did we see a dec.point already?
		SCF
		RET	NZ		;yes -> exit with Carry
		SET	6,B		;no, set flag
		EXX
AtoR2:	;advance one character
		INC	IX
		JR	AtoR1

AtoR3:		CP	'E'		;exponent?
		JR	Z,AtoR5		;yes -> end of mantissa
	;neither '.' nor 'E', has to be digit
		CALL	CheckDecim	;get char and check for digit
		JR	NC,SExp10	;none -> finished
		EX	AF,AF'		;save digit
		CALL	Mult10		;times 10.0
		RET	C		;Overflow -> exit
		EX	AF,AF'		;get digit
		EXX
		PUSH	BC		;save flags/counter
		LD	L,A		;digit -> integer
		LD	H,0
		CALL	Int2Real	;-> real
		CALL	AddReal		;and add it
		EXX
		POP	BC		;get flags/counter
		RET	C		;Overflow -> exit
		BIT	6,B		;if we are after the dec. point
		JR	Z,AtoR4		;then increase the position of
		DEC	C		;the decimal point
AtoR4:		EXX
		JR	AtoR2		;next char

AtoR5:		;evaluate exponent
		CALL	SExp10		;first multiply mantissa by 10^(-X)
		RET	C		;exit if overflow
		EXX
		SET	4,B		;set flag 'exponent'
		INC	IX		;get next character
		LD	A,(IX)		;look if sign
		CP	'+'
		JR	Z,AtoR6
		CP	'-'
		JR	NZ,AtoR7
		SET	5,B		;set flag 'negative exp.'
AtoR6:		INC	IX
AtoR7:		CALL	GetDecimal	;now we need a digit
		CCF
		RET	C		;error if not
		LD	C,A		;save exp. (first digit)
		INC	IX
		CALL	GetDecimal	;another digit following?
		JR	NC,AtoR8	;no -> exp. finished
		INC	IX
		LD	D,A		;save second digit
		LD	A,C		;first digit times 10
		ADD	A,A
		ADD	A,A
		ADD	A,C
		ADD	A,A
		ADD	A,D		;add second digit
		LD	C,A
AtoR8:		BIT	5,B		;exp. negative?
		JR	Z,AtoR9		;no -> exp. finished
		LD	A,C		;yes, negate
		NEG
		LD	C,A
AtoR9:		EXX			;exp. is in C
;<fall through>

;E: real number X in BCDEHL, exp. E base 10 in C'
;V: AF, BCDEHL, DEHL'
;R: carry if overflow, else X * 10^E in BCDEHL
SExp10:		EXX
		LD	A,C
		ADD	A,80h		;check E
		CP	5Ah		;smaller than -38?
		RET	C		;yes -> overflow
		CP	0A6h		;greater then +38?
		CCF
		RET	C		;yes -> overflow
		PUSH	BC		;save flags/counter
		PUSH	IX		;save pointer to string
		LD	A,C
		CALL	Exp10		;this is where it happens
		POP	IX		;now carry if overflow
		EXX
		POP	BC
		EXX
		RET

;Get char, check for decimal digit and convert to binary
;E: pointer in IX
;V: AF / R: Carry if decimal digit and value in A, else NCarry
GetDecimal:	LD	A,(IX)
;<fall through>

;Check for decimal digit and convert to binary
;E: char in A
;V: AF / R: Carry if decimal digit and value in A, else NCarry
CheckDecim:	SUB	'0'
		CCF
		RET	NC
		CP	10
		RET

;times 10^E
;E: real number X in BCDEHL', exp. E of base 10 in A
;V: AF, BCDEHL, BCDEHL', IX
;R: carry if overflow, else X * 10^E in BCDEHL
;Method:
; (1)	if E positive, then compute 10^E * X
;	else compute X / 10^abs(E)
; (2)	10^E = 10^(E div 4) * 10^(E mod 4)
;	the first part comes from a table, the second one
;	is computed by iterated multiplication by 10.0
;Attention: there is no range check for exp. E!
Exp10:		PUSH	AF		;save E
		OR	A		;E positive?
		JP	P,Exp101
		NEG			;no -> negate
Exp101:		PUSH	AF		;save absolute value of E
		SRL	A		;div 4 (cause table has only entries for
		SRL	A		;every 4th power of 10 to reduce space)
		INC	A		;to treat special case 0 also
		LD	HL,-6
		LD	DE,6		;length of a table entry
Exp102:		ADD	HL,DE
		DEC	A
		JR	NZ,Exp102
		EX	DE,HL		;offset in table to DE
		LD	IX,TensTab
		ADD	IX,DE		;IX points to desired entry
		CALL	GetReal		;get entry
		POP	AF		;get absolute value of E
		AND	3		;mod 4
		JR	Z,Exp104	;=0? yes -> finished
Exp103:		PUSH	AF		;no, save counter
		CALL	Mult10		;times 10.0
		POP	AF		;no overflow can occurr here,
		DEC	A		;because max. 10^39
		JR	NZ,Exp103
Exp104:	;now 10^abs(E) in BCDEHL, X still in BCDEHL'
		POP	AF		;fetch E
		OR	A		;positive?
		JP	P,@MultReal	;yes -> multiplicate
		EXX			;no, divide X by 10^abs(E)
		JP	@DivReal
	;here we have to care about overflow when multiplying/dividing

TensTab:	;table with powers of 10
	DEFB	081h,000h,000h,000h,000h,000h	;= 1.0000000000E+00 = 10^0
	DEFB	08Eh,000h,000h,000h,040h,01Ch	;= 1.0000000000E+04 = 10^4
	DEFB	09Bh,000h,000h,020h,0BCh,03Eh	;= 1.0000000000E+08 = 10^8
	DEFB	0A8h,000h,010h,0A5h,0D4h,068h	;= 1.0000000000E+12 = 10^12
	DEFB	0B6h,004h,0BFh,0C9h,01Bh,00Eh	;= 1.0000000000E+16 = 10^16
	DEFB	0C3h,0ACh,0C5h,0EBh,078h,02Dh	;= 1.0000000000E+20 = 10^20
	DEFB	0D0h,0CDh,0CEh,01Bh,0C2h,053h	;= 1.0000000000E+24 = 10^24
	DEFB	0DEh,0F9h,078h,039h,03Fh,001h	;= 1.0000000000E+28 = 10^28
	DEFB	0EBh,02Bh,0A8h,0ADh,0C5h,01Dh	;= 1.0000000000E+32 = 10^32
	DEFB	0F8h,0C9h,07Bh,0CEh,097h,040h	;= 1.0000000000E+36 = 10^36

;times 10.0
;E: positive real number X in BCDEHL
;V: AF, BCDEHL
;R: Carry if overflow
;   X times 10 in BCDEHL
;Method:
;   X := 8 * (X + X/4)
Mult10:		LD	A,L		;X=0? Check for special case zero.
		OR	A
		RET	Z		;yes -> result is zero too
		SET	7,B		;complete mantissa
		PUSH	BC		;save mantissa
		PUSH	DE
		LD	A,H
		CALL	SRLMantisse	;div. mantissa by 4
		CALL	SRLMantisse
		ADD	A,H		;add saved mantissa onto this
		LD	H,A
		EX	(SP),HL
		ADC	HL,DE
		EX	DE,HL
		POP	HL
		EX	(SP),HL
		ADC	HL,BC
		LD	B,H
		LD	C,L
		POP	HL
		JR	NC,Mult101	;norm mantissa
		CALL	RRMantisse	;shift if carry
		INC	L		;adapt characteristic
		SCF
		RET	Z		;if overflow
Mult101:	LD	A,L		;times 8
		ADD	A,3		;produces carry if overflow
		LD	L,A
		RES	7,B		;sign: always positive
		RET

;SET Arithmetic
;--------------
.COMMENT %----------------------------------------------------------------
Basics:

In Turbo Pascal a SET is an array of 256 bits, where each bit corresponds
to an element of the base set. An element is present iff the correspon-
ding bit is set. Thus the base set may only have a maximum of 256 elements.
Furthermore it is fixed that the first bit corresponds to an element that
has a value or order of 0, whereas the last bit of the set corresponds to
the element that has a value or order of 255. Base sets like -100..-10
or 2100..2200 are thus not accepted in spite of the fact that they have
less than 256 elements.

Internally a SET is always 32 byte in size and always kept on the stack.
SET variables in memory on the other hand only allocate the minumum
size needed to store the smallest and biggest element of the base set.
They can have 1 to 32 bytes. They are a byte-oriented clippings of the
internal representation; i.e. SizeOf(SET OF 7..8) is 2 not 1 because
it crosses a byte-boundary. The actual size is fixed no matter what
elements are in the SET or not, so the compiler has no need to store
the actual size together with the SET. The size information is used
by the compiler to produce code for accessing these variables accordingly.
Structure: the element with value or order 0 is stored in bit 0 of
the first byte. The 7th element is in bit 7 of byte 0, the 8th in bit 0
of the second byte, and so on.
----------------------------------------------------------------% ;COMMENT

;NEqual SET
;E: SETs X, Y on stack
;V: AF, BC, DE, HL, IX, IY, SP
;R: HL=1 (TRUE) if X <> Y, else HL=0 (FALSE)
NESet:		LD	C,1		;TRUE (not equal)
		JR	EQSet1

;EQual SET
;E: SETs X, Y on stack
;V: AF, BC, DE, HL, IX, IY, SP
;R: HL=1 (TRUE) if X = Y, else HL=0 (FALSE)
EQSet:		LD	C,0		;FALSE (not equal)
EQSet1:		CALL	GetSetPtr	;get pointers to SETs
EQSet2:		LD	A,(DE)		;compare
		CP	(HL)
		JR	NZ,SRetBool
		INC	HL
		INC	DE
		DJNZ	EQSet2
		LD	A,C		;equal, turn around
		XOR	1		;logic value
		LD	C,A

SRetBool: ;result is in C, return address in IX
		LD	HL,64		;throw away both SETs
		ADD	HL,SP
		LD	SP,HL
		LD	L,C		;extend result to word
		LD	H,0
		JP	(IX)

;Less or Equal SET (subset)
;E: SETs X, Y on stack
;V: AF, BC, DE, HL, IX, IY, SP
;R: HL=1 (TRUE) if X subset of Y, else HL=0 (FALSE)
LESet:		LD	C,1
		JR	GESet1

;Greater or Equal SET (subset)
;E: SETs X, Y on stack
;V: AF, BC, DE, HL, IX, IY, SP
;R: HL=1 (TRUE) if Y subset of X, else HL=0 (FALSE)
GESet:		LD	C,0
GESet1:		CALL	GetSetPtr	;get pointers to SETs
		DEC	C		;Greater or Equal?
		JR	NZ,GESet2	;yes -> result ok
		EX	DE,HL		;no, exchange X and Y
GESet2:		LD	C,0		;suppose FALSE
GESet3:		LD	A,(DE)		;look, if X = (X OR Y)
		OR	(HL)		;if so, then Y subset of X
		CP	(HL)
		JR	NZ,SRetBool
		INC	HL
		INC	DE
		DJNZ	GESet3
		LD	C,1		;return TRUE
		JR	SRetBool

;SET Addition (union)
;E: SETs X, Y on stack
;V: AF, B, DE, HL, IX, IY, SP
;R: X+Y on stack
AddSet:		CALL	GetSetPtr	;get pointers to SETs
AddSet1:	LD	A,(DE)		;X := X OR Y
		OR	(HL)
		LD	(HL),A
		INC	HL
		INC	DE
		DJNZ	AddSet1

SRetSet: ;DE points to SET X (behind SET Y), return address in IX
		EX	DE,HL		;result is in SET X,
		LD	SP,HL		;so throw away SET Y
		JP	(IX)

;SET Subtraction (set difference)
;E: SETs X, Y on stack
;V: AF, B, DE, HL, IX, IY, SP
;R: X-Y on stack
SubSet:		CALL	GetSetPtr
SubSet1:	LD	A,(DE)		;X := X AND (NOT Y)
		CPL
		AND	(HL)
		LD	(HL),A
		INC	HL
		INC	DE
		DJNZ	SubSet1
		JR	SRetSet

;SET Multiplication (damn! my dictionary does not contain that word.
;I explain: A set whose elements are in both source sets)
;E: SETs X, Y on stack
;V: AF, B, DE, HL, IX, IY, SP
;R: X*Y on stack
MultSet:	CALL	GetSetPtr
MultSet1:	LD	A,(DE)		;X := X AND Y
		AND	(HL)
		LD	(HL),A
		INC	HL
		INC	DE
		DJNZ	MultSet1
		JR	SRetSet

;E: sets X, Y on stack (under two return addresses)
;V: F, B, DE, HL, IX, IY, SP
;R: pointer to X in HL, pointer to Y in DE
;   B=32, second return address in IX
GetSetPtr:	POP	IY		;return address 1
		POP	IX		;return address 2
		LD	HL,0
		ADD	HL,SP
		EX	DE,HL		;pointer to first SET to DE
		LD	HL,32		;length of a SET
		LD	B,L		;initialise counter
		ADD	HL,SP		;pointer to second SET to HL
		JP	(IY)

;element IN SET
;E: SET X on stack, element E on stack (word-wide)
;V: AF, B, HL, IX, SP
InSet:		POP	IX		;return address
		LD	HL,21h
		ADD	HL,SP		;compute pointer to HighByte of E
		LD	A,(HL)		;get Highbyte of E
		OR	A		;=0?
		JR	Z,InSet1	;yes -> look if there
		XOR	A		;no, we don't have that many elements,
		JR	InSet2		;so we can't find it

InSet1:		DEC	HL		;pointer to LowByte of E
		LD	B,(HL)		;get it
		CALL	PtrMskSet	;compute pointer and mask
		AND	(HL)		;is it in set?
InSet2:	;result now in zero flag
		LD	HL,22h		;throw away SET and element
		ADD	HL,SP
		LD	SP,HL
		LD	HL,0		;suppose FALSE
		JR	Z,InSet3	;was it false? yes -> finished
		INC	HL		;no, change to TRUE
InSet3:		JP	(IX)

;END OF TP-RTL5.INC
;------------------
